from __future__ import print_function

from spinnman import exceptions

import sys
import traceback


class AbstractProcess(object):
    """ An abstract process for talking to SpiNNaker efficiently
    """

    def __init__(self):
        self._exception = None
        self._traceback = None
        self._error_request = None

    def _receive_error(self, request, exception, tb):
        self._error_request = request
        self._exception = exception
        self._traceback = tb

    def is_error(self):
        return self._exception is not None

    def check_for_error(self, print_exception=False):
        if self._exception is not None:
            if print_exception:
                sdp_header = self._error_request.sdp_header
                print("Error in request to {}, {}, {}".format(
                    sdp_header.destination_chip_x,
                    sdp_header.destination_chip_y,
                    sdp_header.destination_cpu),
                    file=sys.stderr)
                for line in traceback.format_exception_only(
                        self._exception.__class__, self._exception):
                    print(line, end="", file=sys.stderr)
                for line in traceback.format_tb(self._traceback):
                    print(line, end="", file=sys.stderr)

            exc_info = sys.exc_info()
            sdp_header = self._error_request.sdp_header
            self._exception = exceptions.SpinnmanGenericProcessException(
                self._exception, exc_info[2], sdp_header.destination_chip_x,
                sdp_header.destination_chip_y, sdp_header.destination_cpu)
            raise self._exception, None, self._traceback



from spinnman.exceptions import SpinnmanInvalidParameterException
from spinnman.messages.spinnaker_boot.spinnaker_boot_op_code \
    import SpinnakerBootOpCode

import struct

BOOT_MESSAGE_VERSION = 1


class SpinnakerBootMessage(object):
    """ A message used for booting the board
    """

    def __init__(self, opcode, operand_1, operand_2, operand_3, data=None,
                 offset=0):
        """
        :param opcode: The operation of this packet
        :type opcode:\
                    :py:class:`spinnman.messages.spinnaker_boot.spinnaker_boot_op_code.SpinnakerBootOpCode`
        :param operand_1: The first operand
        :type operand_1: int
        :param operand_2: The second operand
        :type operand_2: int
        :param operand_3: The third operand
        :type operand_3: int
        :param data: The optional data, up to 256 words
        :type data: str
        :param offset: The offset of the valid data
        :type offset: int
        :raise spinnman.exceptions.SpinnmanInvalidParameterException: If the\
                    opcode is not a valid value
        """
        if data is not None and len(data) > (256 * 4):
            raise SpinnmanInvalidParameterException(
                "len(data)", str(len(data)),
                "A boot packet can contain at most 256 words of data")

        self._opcode = opcode
        self._operand_1 = operand_1
        self._operand_2 = operand_2
        self._operand_3 = operand_3
        self._data = data
        self._offset = offset

    @property
    def opcode(self):
        """ The operation of this packet

        :return: The operation code
        :rtype:\
                :py:class:`spinnman.messages.spinnaker_boot.spinnaker_boot_op_code.SpinnakerBootOpCode`
        """
        return self._opcode

    @property
    def operand_1(self):
        """ The first operand

        :return: The operand
        :rtype: int
        """
        return self._operand_1

    @property
    def operand_2(self):
        """ The second operand

        :return: The second operand
        :rtype: int
        """
        return self._operand_2

    @property
    def operand_3(self):
        """ The third operand

        :return: The third operand
        :rtype: int
        """
        return self._operand_3

    @property
    def data(self):
        """ The data

        :return: The data or None if no data
        :rtype: bytearray
        """
        return self._data

    @property
    def bytestring(self):
        """ The message as a bytestring
        """
        data = ""
        if self._data is not None:
            data = self._data[self._offset:]
        return struct.pack(
            ">HIIII", BOOT_MESSAGE_VERSION, self._opcode.value,
            self._operand_1, self._operand_2, self._operand_3) + data

    @staticmethod
    def from_bytestring(data, offset):
        (opcode_value, operand_1, operand_2, operand_3) = struct.unpack_from(
            "2xIIII", data, offset)
        the_data = None
        if len(data) - offset > 0:
            the_data = data
        return SpinnakerBootMessage(
            SpinnakerBootOpCode(opcode_value), operand_1, operand_2, operand_3,
            the_data, offset + 18)



from collections import OrderedDict


class CPUInfos(object):
    """ A set of CPU information objects
    """

    def __init__(self):
        self._cpu_infos = OrderedDict()

    def add_processor(self, x, y, processor_id, cpu_info):
        """ Add a processor on a given chip to the set

        :param x: The x-coordinate of the chip
        :type x: int
        :param y: The y-coordinate of the chip
        :type y: int
        :param processor_id: A processor id
        :type processor_id: int
        :param cpu_info: The CPU information for the core
        :type cpu_info: :py:class:`spinnman.model.cpu_info.CPUInfo`
        """
        self._cpu_infos[x, y, processor_id] = cpu_info

    @property
    def cpu_infos(self):
        """ the one per core core info

        :return: iterable of x,y,p core info
        """
        return self._cpu_infos.iteritems()

    def __iter__(self):
        return self._cpu_infos.__iter__()

    def iteritems(self):
        """ Get an iterable of (x, y, p), cpu_info
        """
        return self._cpu_infos.iteritems()

    def itervalues(self):
        """ Get an iterable of cpu_info
        """
        return self._cpu_infos.itervalues()

    def iterkeys(self):
        """ Get an iterable of (x, y, p)
        """
        return self._cpu_infos.iterkeys()

    def __len__(self):
        """ The total number of processors that are in these core subsets
        """
        return len(self._cpu_infos)



import os
import time
import logging

logger = logging.getLogger(__name__)


def generate_machine_report(report_directory, machine, connections):
    """Generate report on the physical structure of the target SpiNNaker \
    machine.

   :param report_directory: the directory to which reports are stored
   :param machine: the machine python object
   :param connections: the list of connections to the machine
   :type report_directory: str
   :type machine: spinnmachine.machine.Machine object
   :type connections: iterable of implementations of \
               spinnman.connections.abstract_connection.AbstractConnection
   :rtype: None
   :raise IOError: when a file cannot be opened for some reason
   """
    file_name = report_directory + os.sep + "machine_structure.rpt"
    f_machine_struct = None
    try:
        f_machine_struct = open(file_name, "w")
    except IOError:
        logger.error("Generate_placement_reports: Can't open file {} for "
                     "writing.".format(file_name))
    f_machine_struct.write("\t\tTarget SpiNNaker Machine Structure\n")
    f_machine_struct.write("\t\t==================================\n\n")
    time_date_string = time.strftime("%c")
    f_machine_struct.write("Generated: %s" % time_date_string)
    f_machine_struct.write(" for target machine '{}'".format(connections))
    f_machine_struct.write("\n\n")

    x_dim = machine.max_chip_x + 1
    y_dim = machine.max_chip_y + 1
    f_machine_struct.write("Machine dimensions (in chips) x : {}  y : {}\n\n"
                           .format(x_dim, y_dim))

    # TODO: Add further details on the target machine.
    f_machine_struct.write("\t\tMachine router information\n")
    f_machine_struct.write("\t\t==========================\n\n")
    for x in range(machine.max_chip_x + 1):
        for y in range(machine.max_chip_y + 1):
            chip = machine.get_chip_at(x, y)
            if chip:
                f_machine_struct.write("Information for chip {}:{}\n"
                                       .format(chip.x, chip.y))
                f_machine_struct.write(
                    "Neighbouring chips \n{}\n"
                    .format(chip.router.get_neighbouring_chips_coords()))
                f_machine_struct.write("Router list of links for this chip"
                                       " are: \n")
                for link in chip.router.links:
                    f_machine_struct.write("\t{}\n".format(link))
                f_machine_struct.write("\t\t==========================\n\n")
    # Close file:
    f_machine_struct.close()



from six import add_metaclass

from spinn_utilities.abstract_base import AbstractBase, abstractmethod
from spinnman.connections.abstract_classes.abstract_connection \
    import AbstractConnection


@add_metaclass(AbstractBase)
class AbstractSDPSender(AbstractConnection):
    """ A sender of SDP messages
    """

    __slots__ = ()

    @abstractmethod
    def send_sdp_message(self, sdp_message):
        """ Sends an SDP message down this connection

        :param sdp_message: The SDP message to be sent
        :type sdp_message: spinnman.messages.sdp.sdp_message.SDPMessage
        :return: Nothing is returned
        :rtype: None
        :raise spinnman.exceptions.SpinnmanIOException: If there is an error\
                    sending the message
        """



from spinnman.messages.scp.impl.scp_read_memory_request \
    import SCPReadMemoryRequest
from spinn_machine.multicast_routing_entry import MulticastRoutingEntry
from spinnman.processes.abstract_multi_connection_process \
    import AbstractMultiConnectionProcess
from spinnman import constants

import functools
import struct

# There are 1024 entries in a routing table
_N_ENTRIES = 1024

# 16 entries fit in a 256-byte read
_ENTRIES_PER_READ = 16

# 64 reads of 16 entries are required for 1024 entries
_N_READS = 64


class GetMultiCastRoutesProcess(AbstractMultiConnectionProcess):

    def __init__(self, connection_selector, app_id=None):
        AbstractMultiConnectionProcess.__init__(self, connection_selector)
        self._entries = [None] * _N_ENTRIES
        self._app_id = app_id

    def handle_read_response(self, offset, response):
        for route_no in range(_ENTRIES_PER_READ):
            (app_id, route, key, mask) = struct.unpack_from(
                "<2xBxIII", response.data, response.offset + (route_no * 16))
            if route < 0xFF000000 and (self._app_id is None or
                                       self._app_id == app_id):
                processor_ids = list()
                for processor_id in range(0, 26):
                    if (route & (1 << (6 + processor_id))) != 0:
                        processor_ids.append(processor_id)
                link_ids = list()
                for link_id in range(0, 6):
                    if (route & (1 << link_id)) != 0:
                        link_ids.append(link_id)
                self._entries[route_no + offset] = MulticastRoutingEntry(
                    key, mask, processor_ids, link_ids, False)

    def get_routes(self, x, y, base_address):

        # Create the read requests
        offset = 0
        for _ in range(_N_READS):
            self._send_request(
                SCPReadMemoryRequest(x, y, base_address + (offset * 16),
                                     constants.UDP_MESSAGE_MAX_SIZE),
                functools.partial(self.handle_read_response, offset))
            offset += _ENTRIES_PER_READ
        self._finish()
        self.check_for_error()

        return [entry for entry in self._entries if entry is not None]



from spinnman.processes.abstract_process import AbstractProcess
from spinnman.connections.scp_request_pipeline import SCPRequestPipeLine


class AbstractMultiConnectionProcess(AbstractProcess):
    """ A process that uses multiple connections in communication
    """

    def __init__(self, next_connection_selector,
                 n_retries=3, timeout=0.5, n_channels=1,
                 intermediate_channel_waits=0):
        AbstractProcess.__init__(self)
        self._scp_request_pipe_lines = dict()
        self._n_retries = n_retries
        self._timeout = timeout
        self._n_channels = n_channels
        self._intermediate_channel_waits = intermediate_channel_waits
        self._next_connection_selector = next_connection_selector

    def _send_request(self, request, callback=None, error_callback=None):
        if error_callback is None:
            error_callback = self._receive_error
        connection = self._next_connection_selector.get_next_connection(
            request)
        if connection not in self._scp_request_pipe_lines:
            scp_request_set = SCPRequestPipeLine(
                connection, n_retries=self._n_retries,
                packet_timeout=self._timeout,
                n_channels=self._n_channels,
                intermediate_channel_waits=self._intermediate_channel_waits)
            self._scp_request_pipe_lines[connection] = scp_request_set
        self._scp_request_pipe_lines[connection].send_request(
            request, callback, error_callback)

    def _finish(self):
        for request_pipeline in self._scp_request_pipe_lines.itervalues():
            request_pipeline.finish()



from enum import Enum


class SCPBMPInfoType(Enum):
    """ The SCP BMP Information Types
    """

    SERIAL = (0, "Serial information")
    CAN_STATUS = (2, "CAN status information")
    ADC = (3, "ADC information")
    IP_ADDR = (4, "IP Address")

    def __new__(cls, value, doc=""):
        obj = object.__new__(cls)
        obj._value_ = value
        return obj

    def __init__(self, value, doc=""):
        self._value_ = value
        self.__doc__ = doc



from spinnman.messages.eieio.abstract_messages.abstract_eieio_message\
    import AbstractEIEIOMessage


class EIEIOCommandMessage(AbstractEIEIOMessage):
    """ An EIEIO command message
    """

    def __init__(self, eieio_command_header, data=None, offset=0):
        """

        :param eieio_command_header: The header of the message
        :type eieio_command_header:\
                    :py:class:`spinnman.messages.eieio.command_messages.eieio_command_header.EIEIOCommandHeader`
        :param data: Optional incoming data
        :type data: str
        :param offset: Offset into the data where valid data begins
        :type offset: int
        """
        AbstractEIEIOMessage.__init__(self)

        # The header
        self._eieio_command_header = eieio_command_header

        # The data
        self._data = data
        self._offset = offset

    @property
    def eieio_header(self):
        return self._eieio_command_header

    @property
    def data(self):
        return self._data

    @property
    def offset(self):
        return self._offset

    @staticmethod
    def from_bytestring(command_header, data, offset):
        return EIEIOCommandMessage(command_header, data, offset)

    @property
    def bytestring(self):
        return self._eieio_command_header.bytestring

    @staticmethod
    def get_min_packet_length():
        return 2

    def __str__(self):
        return "EIEIOCommandMessage:{}".format(self._eieio_command_header)

    def __repr__(self):
        return self.__str__()



from spinnman.messages.scp.abstract_messages.abstract_scp_request\
    import AbstractSCPRequest
from spinnman.messages.sdp.sdp_header import SDPHeader
from spinnman.messages.sdp.sdp_flag import SDPFlag
from spinnman.messages.scp.scp_request_header import SCPRequestHeader
from spinnman.messages.scp.enums.scp_command import SCPCommand
from spinnman.messages.scp.impl.scp_check_ok_response import SCPCheckOKResponse

_IPTAG_SET = 1


class SCPIPTagSetRequest(AbstractSCPRequest):
    """ An SCP Request to set an IP Tag
    """

    def __init__(self, x, y, host, port, tag, strip):
        """
        :param x: The x-coordinate of a chip, between 0 and 255
        :type x: int
        :param y: The y-coordinate of a chip, between 0 and 255
        :type y: int
        :param host: The host address, as an array of 4 bytes
        :type host: bytearray
        :param port: The port, between 0 and 65535
        :type port: int
        :param tag: The tag, between 0 and 7
        :type tag: int
        :param strip: if the SDP header should be striped from the packet.
        :type strip: bool
        """
        strip_value = 0
        if strip:
            strip_value = 1

        super(SCPIPTagSetRequest, self).__init__(
            SDPHeader(
                flags=SDPFlag.REPLY_EXPECTED, destination_port=0,
                destination_cpu=0, destination_chip_x=x,
                destination_chip_y=y),
            SCPRequestHeader(command=SCPCommand.CMD_IPTAG),
            argument_1=(strip_value << 28) | (_IPTAG_SET << 16) | tag,
            argument_2=port,
            argument_3=((host[3] << 24) | (host[2] << 16) |
                        (host[1] << 8) | host[0]))

    def get_scp_response(self):
        return SCPCheckOKResponse("Set IP Tag", "CMD_IPTAG")



from enum import Enum


class SCPIPTagCommand(Enum):
    """ SCP IPTag Commands
    """
    NEW = 0
    SET = 1
    GET = 2
    CLR = 3
    TTO = 4

    def __new__(cls, value, doc=""):
        obj = object.__new__(cls)
        obj._value_ = value
        return obj

    def __init__(self, value, doc=""):
        self._value_ = value
        self.__doc__ = doc



from enum import Enum


class SCPDPRIPacketTypeFlags(Enum):
    """ SCP Dropped Packet Reinjection Packet type flags
    """
    MULTICAST = 1
    POINT_TO_POINT = 2
    NEAREST_NEIGHBOUR = 4
    FIXED_ROUTE = 8

    def __new__(cls, value, doc=""):
        obj = object.__new__(cls)
        obj._value_ = value
        return obj

    def __init__(self, value, doc=""):
        self._value_ = value
        self.__doc__ = doc



"""
AbstractSCPBMPResponse
"""

# spinnman imports
from spinnman.messages.scp.abstract_messages.abstract_scp_response import \
    AbstractSCPResponse


class AbstractSCPBMPResponse(AbstractSCPResponse):
    """
    represents a scp request thats tialored for the bmp connection
    """

    __slots__ = ()

    def __init__(self):
        AbstractSCPResponse.__init__(self)



class MachineDimensions(object):
    """ Represents the size of a machine in chips
    """

    def __init__(self, width, height):
        """
        :param width: The width of the machine in chips
        :type width: int
        :param height: The height of the machine in chips
        :type height: int
        :raise None: No known exceptions are raised
        """
        self._width = width
        self._height = height

    @property
    def width(self):
        """ The width of the machine in chips

        :return: The width
        :rtype: int
        """
        return self._width

    @property
    def height(self):
        """ The height of the machine in chips

        :return: The height
        :rtype: int
        """
        return self._height



import time


class TokenBucket(object):
    """ An implementation of the token bucket algorithm.

    >>> bucket = TokenBucket(80, 0.5)
    >>> print bucket.consume(10)
    True
    Not thread safe.
    """

    __slots__ = ['_capacity', '_tokens', '_fill_rate', '_timestamp']

    def __init__(self, tokens, fill_rate):
        """

        :param tokens: the total tokens in the bucket
        :param fill_rate:\
            the rate in tokens/second that the bucket will be refilled.
        """
        self._capacity = float(tokens)
        self._tokens = float(tokens)
        self._fill_rate = float(fill_rate)
        self._timestamp = time.time()

    def consume(self, tokens, block=True):
        """ Consume tokens from the bucket. Returns True if there were\
            sufficient tokens.

        If there are not enough tokens and block is True, sleeps until the
        bucket is replenished enough to satisfy the deficiency.

        If there are not enough tokens and block is False, returns False.

        It is an error to consume more tokens than the bucket _capacity.

        """
        while block and tokens > self.tokens:
            deficit = tokens - self._tokens
            delay = deficit / self._fill_rate
            time.sleep(delay)

        if tokens <= self.tokens:
            self._tokens -= tokens
            return True
        return False

    @property
    def tokens(self):
        if self._tokens < self._capacity:
            now = time.time()
            delta = self._fill_rate * (now - self._timestamp)
            self._tokens = min(self._capacity, self._tokens + delta)
            self._timestamp = now
        return self._tokens



from six import add_metaclass

from spinn_utilities.abstract_base import AbstractBase, abstractmethod


@add_metaclass(AbstractBase)
class AbstractEIEIODataElement(object):
    """ A marker interface for possible data elements in the EIEIO data packet
    """

    __slots__ = ()

    @abstractmethod
    def get_bytestring(self, eieio_type):
        """ Get a bytestring for the given type

        :param eieio_type: The type of the message being written
        :type eieio_type:\
                    :py:class:`spinnman.messages.eieio.eieio_type.EIEIOType`
        :return: A bytestring for the element
        :rtype: str
        :raise SpinnmanInvalidParameterException: If the type is incompatible\
                    with the element
        """




# spinnman imports
from spinnman import constants
import struct

from spinnman.connections.udp_packet_connections.udp_connection import \
    UDPConnection
from spinnman.messages.scp.enums.scp_result import SCPResult
from spinnman.connections.udp_packet_connections import udp_utils
from spinnman.connections.abstract_classes.abstract_scp_receiver \
    import AbstractSCPReceiver
from spinnman.connections.abstract_classes.abstract_scp_sender \
    import AbstractSCPSender


class UDPBMPConnection(
        UDPConnection, AbstractSCPReceiver, AbstractSCPSender):
    """ A BMP connection which supports queries to the BMP of a SpiNNaker\
        machine
    """

    def __init__(self, cabinet, frame, boards, local_host=None,
                 local_port=None, remote_host=None,
                 remote_port=None):
        """
        :param cabinet: The cabinet number of the connection
        :type cabinet: int
        :param frame: The frame number of the connection
        :type frame: int
        :param boards: The boards that the connection can control on the same\
                backplane
        :type boards: iterable of int
        :param local_host: The optional ip address or host name of the local\
                interface to listen on
        :type local_host: str
        :param local_port: The optional local port to listen on
        :type local_port: int
        :param remote_host: The optional remote host name or ip address to\
                send messages to.  If not specified, sending will not be\
                possible using this connection
        :type remote_host: str
        :param remote_port: The optional remote port number to send messages\
                to.  If not specified, sending will not be possible using this\
                connection
        """
        if remote_port is None:
            remote_port = constants.SCP_SCAMP_PORT
        UDPConnection.__init__(
            self, local_host, local_port, remote_host, remote_port)
        AbstractSCPReceiver.__init__(self)
        AbstractSCPSender.__init__(self)
        self._cabinet = cabinet
        self._frame = frame
        self._boards = boards

    @property
    def cabinet(self):
        """ The cabinet id of the BMP

        :rtype: int
        """
        return self._cabinet

    @property
    def frame(self):
        """ The frame id of the BMP

        :rtype: int
        """
        return self._frame

    @property
    def boards(self):
        """ The set of boards supported by the BMP

        :rtype: iterable of int
        """
        return self._boards

    @property
    def chip_x(self):
        """ Defined to satisfy the AbstractSCPSender - always 0 for a BMP
        """
        return 0

    @property
    def chip_y(self):
        """ Defined to satisfy the AbstractSCPSender - always 0 for a BMP
        """
        return 0

    def get_scp_data(self, scp_request):
        udp_utils.update_sdp_header_for_udp_send(scp_request.sdp_header, 0, 0)
        return struct.pack("<2x") + scp_request.bytestring

    def receive_scp_response(self, timeout=1.0):
        data = self.receive(timeout)
        result, sequence = struct.unpack_from("<2H", data, 10)
        return SCPResult(result), sequence, data, 2

    def send_scp_request(self, scp_request):
        self.send(self.get_scp_data(scp_request))

    def __repr__(self):
        return \
            "UDPBMPConnection(cabinet={}, frame={}, boards={}, local_host={},"\
            "local_port={}, remote_host={}, remote_port={}".format(
                self._cabinet, self._frame, self._boards,
                self.local_ip_address, self.local_port, self.remote_ip_address,
                self.remote_port)



from spinnman.connections.udp_packet_connections.udp_connection\
    import UDPConnection
from spinnman.connections.abstract_classes.abstract_eieio_receiver\
    import AbstractEIEIOReceiver
from spinnman.connections.abstract_classes.abstract_eieio_sender\
    import AbstractEIEIOSender
from spinnman.connections.abstract_classes.abstract_listenable\
    import AbstractListenable
from spinnman.messages.eieio.create_eieio_command \
    import read_eieio_command_message
from spinnman.messages.eieio.create_eieio_data import read_eieio_data_message

import struct


class UDPEIEIOConnection(UDPConnection, AbstractEIEIOReceiver,
                         AbstractEIEIOSender, AbstractListenable):
    """ A UDP connection for sending and receiving raw EIEIO messages
    """

    def __init__(self, local_host=None, local_port=None, remote_host=None,
                 remote_port=None):
        """

        :param local_host: The optional ip address or host name of the local\
                interface to listen on
        :type local_host: str
        :param local_port: The optional local port to listen on
        :type local_port: int
        :param remote_host: The optional remote host name or ip address to\
                send messages to.  If not specified, sending will not be\
                possible using this connection
        :type remote_host: str
        :param remote_port: The optional remote port number to send messages\
                to.  If not specified, sending will not be possible using this\
                connection
        """
        UDPConnection.__init__(
            self, local_host, local_port, remote_host, remote_port)
        AbstractEIEIOReceiver.__init__(self)
        AbstractEIEIOSender.__init__(self)
        AbstractListenable.__init__(self)

    def receive_eieio_message(self, timeout=None):
        data = self.receive(timeout)
        header = struct.unpack_from("<H", data)[0]
        if header & 0xC000 == 0x4000:
            eieio_message = read_eieio_command_message(data, 0)
        else:
            eieio_message = read_eieio_data_message(data, 0)

        return eieio_message

    def send_eieio_message(self, eieio_message):
        self.send(eieio_message.bytestring)

    def send_eieio_message_to(self, eieio_message, ip_address, port):
        self.send_to(eieio_message.bytestring, (ip_address, port))

    def get_receive_method(self):
        return self.receive_eieio_message

    def __repr__(self):
        return \
            "UDPEIEIOConnection(local_host={}, local_port={}," \
            "remote_host={}, remote_port={})".format(
                self.local_ip_address, self.local_port,
                self.remote_ip_address, self.remote_port)



from enum import Enum


class SpinnakerBootOpCode(Enum):
    """ Boot message Operation Codes
    """

    HELLO = 0x41
    FLOOD_FILL_START = 0x1
    FLOOD_FILL_BLOCK = 0x3
    FLOOD_FILL_CONTROL = 0x5

    def __new__(cls, value, doc=""):
        obj = object.__new__(cls)
        obj._value_ = value
        return obj

    def __init__(self, value, doc=""):
        self._value_ = value
        self.__doc__ = doc



from threading import Thread
import traceback
from multiprocessing.pool import ThreadPool


class ConnectionListener(Thread):
    """ Listens to a connection and calls callbacks with new messages when \
        they arrive
    """

    def __init__(self, connection, n_processes=4):
        """

        :param connection: An AbstractListenable connection to listen to
        :param n_processes: The number of threads to use when calling\
                callbacks
        """
        Thread.__init__(
            self,
            name="Connection listener for connection {}".format(connection))
        self._connection = connection
        self._get_message_call = connection.get_receive_method()
        self._callback_pool = ThreadPool(processes=n_processes)
        self._done = False
        self._callbacks = set()
        self.setDaemon(True)

    def run(self):
        while not self._done:
            try:
                if self._connection.is_ready_to_receive(timeout=1):
                    message = self._get_message_call()
                    for callback in self._callbacks:
                        self._callback_pool.apply_async(callback, [message])
            except:
                if not self._done:
                    traceback.print_exc()
        self._callback_pool.close()
        self._callback_pool.join()

    def add_callback(self, callback):
        """ Add a callback to be called when a message is received

        :param callback: A callable which takes a single parameter, which is\
                the message received
        """
        self._callbacks.add(callback)

    def close(self):
        """ Closes the listener.  Note that this does not close the provider\
            of the messages; this instead marks the listener as closed.  The\
            listener will not truly stop until the get message call returns.
        """
        self._done = True
        self.join()



from spinnman.exceptions import SpinnmanInvalidPacketException
from spinnman.messages.eieio.data_messages.eieio_key_payload_data_element\
    import EIEIOKeyPayloadDataElement
from spinnman.messages.eieio.data_messages.eieio_key_data_element\
    import EIEIOKeyDataElement
from spinnman.messages.eieio.data_messages.eieio_data_header\
    import EIEIODataHeader
from spinnman.messages.eieio.abstract_messages.abstract_eieio_message\
    import AbstractEIEIOMessage
from spinnman.messages.eieio.eieio_type import EIEIOType
from spinnman.messages.eieio.eieio_prefix import EIEIOPrefix
from spinnman import constants

import math
import struct


class EIEIODataMessage(AbstractEIEIOMessage):
    """ An EIEIO Data message
    """

    def __init__(self, eieio_header, data=None, offset=0):
        """

        :param eieio_header: The header of the message
        :type eieio_header:\
                    :py:class:`spinnman.messages.eieio.data_messages.eieio_data_header.EIEIODataHeader`
        :param data: Optional data contained within the packet
        :type data: str
        :param offset: Optional offset where the valid data starts
        :type offset: int
        """

        # The header
        self._eieio_header = eieio_header

        # Elements to be written
        self._elements = b""

        # Keeping track of the reading of the data
        self._data = data
        self._offset = offset
        self._elements_read = 0

    @property
    def eieio_header(self):
        return self._eieio_header

    @staticmethod
    def min_packet_length(eieio_type, is_prefix=False, is_payload_base=False):
        """ The minimum length of a message with the given header, in bytes

        :param eieio_type: the type of message
        :type eieio_type:\
                    :py:class:`spinnman.spinnman.messages.eieio.eieio_type.EIEIOType`
        :param is_prefix: True if there is a prefix, False otherwise
        :type is_prefix: bool
        :param is_payload_base: True if there is a payload base, False\
                    otherwise
        :type is_payload_base: bool
        :return: The minimum size of the packet in bytes
        :rtype: int
        """
        header_size = EIEIODataHeader.get_header_size(eieio_type, is_prefix,
                                                      is_payload_base)
        return header_size + eieio_type.payload_bytes

    @property
    def max_n_elements(self):
        """ The maximum number of elements that can fit in the packet

        :rtype: int
        """
        return int(math.floor((constants.UDP_MESSAGE_MAX_SIZE -
                               self._eieio_header.size) /
                              (self._eieio_header.eieio_type.key_bytes +
                               self._eieio_header.eieio_type.payload_bytes)))

    @property
    def n_elements(self):
        """ The number of elements in the packet
        """
        return self._eieio_header.count

    @property
    def size(self):
        """ The size of the packet with the current contents
        """
        return (self._eieio_header.size +
                ((self._eieio_header.eieio_type.key_bytes +
                 self._eieio_header.eieio_type.payload_bytes) *
                 self._eieio_header.count))

    def add_element(self, element):
        """ Add an element to the message.  The correct type of element must\
            be added, depending on the header values

        :param element: The element to be added
        :type element:\
                    :py:class:`spinnman.messages.eieio.data_messages.abstract_eieio_data_element.AbstractEIEIODataElement`
        :raise SpinnmanInvalidParameterException: If the element is not\
                    compatible with the header
        :raise SpinnmanInvalidPacketException: If the message was created to\
                    read data
        """
        if self._data is not None:
            raise SpinnmanInvalidPacketException(
                "EIEIODataMessage", "This packet is read-only")

        self._elements += element.get_bytestring(self._eieio_header.eieio_type)
        self._eieio_header.increment_count()

    @property
    def is_next_element(self):
        """ Determine if there is another element to be read

        :return: True if the message was created with data, and there are more\
                    elements to be read
        :rtype: bool
        """
        return (self._data is not None and
                self._elements_read < self._eieio_header.count)

    @property
    def next_element(self):
        """ The next element to be read, or None if no more elements.  The\
            exact type of element returned depends on the packet type

        :rtype:\
                    :py:class:`spinnman.messages.eieio.data_messages.abstract_eieio_data_element.AbstractEIEIODataElement`
        """
        if not self.is_next_element:
            return None
        self._elements_read += 1
        key = None
        payload = None
        if self._eieio_header.eieio_type == EIEIOType.KEY_16_BIT:
            key = struct.unpack_from("<H", self._data, self._offset)[0]
            self._offset += 2
        if self._eieio_header.eieio_type == EIEIOType.KEY_32_BIT:
            key = struct.unpack_from("<I", self._data, self._offset)[0]
            self._offset += 4
        if self._eieio_header.eieio_type == EIEIOType.KEY_PAYLOAD_16_BIT:
            key, payload = struct.unpack_from("<HH", self._data, self._offset)
            self._offset += 4
        if self._eieio_header.eieio_type == EIEIOType.KEY_PAYLOAD_32_BIT:
            key, payload = struct.unpack_from("<II", self._data, self._offset)
            self._offset += 8

        if self._eieio_header.prefix is not None:
            if self._eieio_header.prefix_type == EIEIOPrefix.UPPER_HALF_WORD:
                key = key | (self._eieio_header.prefix << 16)
            else:
                key = key | self._eieio_header.prefix

        if self._eieio_header.payload_base is not None:
            if payload is not None:
                payload = payload | self._eieio_header.payload_base
            else:
                payload = self._eieio_header.payload_base

        if payload is None:
            return EIEIOKeyDataElement(key)
        else:
            return EIEIOKeyPayloadDataElement(key, payload,
                                              self._eieio_header.is_time)

    @property
    def bytestring(self):
        return self._eieio_header.bytestring + self._elements

    def __str__(self):
        if self._data is not None:
            return "EIEIODataMessage:{}:{}".format(
                self._eieio_header, self._eieio_header.count)
        return "EIEIODataMessage:{}:{}".format(
            self._eieio_header, self._elements)

    def __repr__(self):
        return self.__str__()



import struct

from spinnman.messages.scp.abstract_messages.abstract_scp_response\
    import AbstractSCPResponse
from spinnman.messages.scp.enums.scp_result import SCPResult
from spinnman.exceptions import SpinnmanUnexpectedResponseCodeException
from spinnman.exceptions import SpinnmanInvalidParameterException


class SCPSDRAMAllocResponse(AbstractSCPResponse):
    """ An SCP response to a request to allocate space in SDRAM
    """

    def __init__(self, size):
        """
        """
        AbstractSCPResponse.__init__(self)
        self._size = size
        self._base_address = None

    def read_data_bytestring(self, data, offset):
        """ See\
            :py:meth:`spinnman.messages.scp.abstract_scp_response.AbstractSCPResponse.read_data_bytestring`
        """
        result = self.scp_response_header.result
        if result != SCPResult.RC_OK:
            raise SpinnmanUnexpectedResponseCodeException(
                "SDRAM Allocation", "CMD_ALLOC", result.name)
        self._base_address = struct.unpack_from("<I", data, offset)[0]

        # check that the base address is not null (0 in python case) as
        # this reflects a issue in the command on spinnaker side
        if self._base_address == 0:
            raise SpinnmanInvalidParameterException(
                "SDRAM Allocation response base address", self._base_address,
                "Could not allocate {} bytes of SDRAM".format(self._size))

    @property
    def base_address(self):
        """ The base address allocated, or 0 if none

        :rtype: int
        """
        return self._base_address



from enum import Enum


class P2PTableRoute(Enum):
    """ P2P Routing table routes
    """
    EAST = 0b000
    NORTH_EAST = 0b001
    NORTH = 0b010
    WEST = 0b011
    SOUTH_WEST = 0b100
    SOUTH = 0b101
    NONE = (0b110, "No route to this chip")
    MONITOR = (0b111, "Route to the monitor on the current chip")

    def __new__(cls, value, doc=""):
        obj = object.__new__(cls)
        obj._value_ = value
        return obj

    def __init__(self, value, doc=""):
        self._value_ = value
        self.__doc__ = doc



_SDP_SOURCE_PORT = 7
_SDP_SOURCE_CPU = 31
_SDP_TAG = 0xFF


def update_sdp_header_for_udp_send(sdp_header, source_x, source_y):
        """ Apply defaults to the sdp header for sending over UDP

        :param sdp_header: The SDP header values
        :type sdp_header:\
                    :py:class:`spinnman.messages.sdp.sdp_header.SDPHeader`
        :return: Nothing is returned
        """
        sdp_header.tag = _SDP_TAG
        sdp_header.source_port = _SDP_SOURCE_PORT
        sdp_header.source_cpu = _SDP_SOURCE_CPU
        sdp_header.source_chip_x = source_x
        sdp_header.source_chip_y = source_y



from spinnman.messages.scp.impl.scp_version_request import SCPVersionRequest
from spinnman.processes.abstract_single_connection_process \
    import AbstractSingleConnectionProcess


class GetVersionProcess(AbstractSingleConnectionProcess):
    """ A process for getting the version of the machine
    """

    def __init__(self, connection_selector):
        AbstractSingleConnectionProcess.__init__(self, connection_selector)
        self._version_info = None

    def _get_response(self, version_response):
        self._version_info = version_response.version_info

    def get_version(self, x, y, p):
        self._send_request(SCPVersionRequest(x=x, y=y, p=p),
                           self._get_response)
        self._finish()

        self.check_for_error()
        return self._version_info



"""
SpinnakerBootMessages
"""

# spinnman imports
from spinnman.messages.spinnaker_boot._system_variables import \
    _system_variable_boot_values as variable_boot_values
from spinnman.messages.spinnaker_boot._system_variables.\
    _system_variable_boot_values import \
    SystemVariableDefinition
from spinnman.messages.spinnaker_boot._system_variables\
    ._system_variable_boot_values import SystemVariableBootValues
from spinnman.messages.spinnaker_boot.spinnaker_boot_message \
    import SpinnakerBootMessage
from spinnman.messages.spinnaker_boot.spinnaker_boot_op_code \
    import SpinnakerBootOpCode

from spinnman.exceptions import SpinnmanInvalidParameterException
from spinnman.exceptions import SpinnmanIOException

# general imports
import os
import math
import time
import array

_BOOT_MESSAGE_DATA_WORDS = 256
_BOOT_MESSAGE_DATA_BYTES = _BOOT_MESSAGE_DATA_WORDS * 4
_BOOT_IMAGE_MAX_BYTES = 32 * 1024
_BOOT_DATA_FILE_NAME = "scamp.boot"
_BOOT_STRUCT_REPLACE_OFFSET = 384 / 4
_BOOT_STRUCT_REPLACE_LENGTH = 128 / 4
_BOOT_DATA_OPERAND_1 = ((_BOOT_MESSAGE_DATA_BYTES / 4) - 1) << 8


class SpinnakerBootMessages(object):
    """ Represents a set of boot messages to be sent to boot the board
    """

    def __init__(self, board_version=None):
        """
        builds the boot messages needed to boot the spinnaker machine

        :param board_version: The version of the board to be booted
        :type board_version: int
        :raise spinnman.exceptions.SpinnmanInvalidParameterException: If the\
                    board version is not supported
        :raise spinnman.exceptions.SpinnmanIOException: If there is an error\
                    assembling the packets
        """
        if (board_version is not None and
                board_version not in
                variable_boot_values.spinnaker_boot_values):
            raise SpinnmanInvalidParameterException(
                "board_version", str(board_version), "Unknown board version")

        # Get the boot packet values
        if board_version is not None:
            spinnaker_boot_value = variable_boot_values.spinnaker_boot_values[
                board_version]
        else:
            spinnaker_boot_value = SystemVariableBootValues()

        current_time = int(time.time())
        spinnaker_boot_value.set_value(
            SystemVariableDefinition.unix_timestamp, current_time)
        spinnaker_boot_value.set_value(
            SystemVariableDefinition.boot_signature, current_time)
        spinnaker_boot_value.set_value(
            SystemVariableDefinition.is_root_chip, 1)

        # Get the data as an array, to be used later
        self._spinnaker_boot_data = array.array(
            "I", spinnaker_boot_value.bytestring)

        # Find the data file and size
        this_dir, _ = os.path.split(__file__)
        boot_data_file_name = os.path.join(
            this_dir, "boot_data", _BOOT_DATA_FILE_NAME)
        boot_data_file_size = os.stat(boot_data_file_name).st_size
        if boot_data_file_size > _BOOT_IMAGE_MAX_BYTES:
            raise SpinnmanIOException(
                "The boot file is too big at {} bytes ("
                "only files up to 32KB are acceptable".format(
                    boot_data_file_size))
        elif boot_data_file_size % 4 != 0:
            raise SpinnmanIOException(
                "The boot file size of {} bytes must"
                " be divisible by 4".format(boot_data_file_size))

        # Compute how many packets to send
        self._boot_data_file = open(boot_data_file_name, "rb")
        self._no_data_packets = int(math.ceil(
            float(boot_data_file_size) / float(_BOOT_MESSAGE_DATA_BYTES)))
        self._n_words_to_read = boot_data_file_size / 4

    def _get_packet_data(self, replace_data=None, offset=None, length=None):

        # Read the next bit of data
        data = array.array("I")
        n_words = min((self._n_words_to_read, _BOOT_MESSAGE_DATA_WORDS))
        self._n_words_to_read -= n_words
        data.fromfile(self._boot_data_file, n_words)

        # If there is any replace_data, replace it now
        if replace_data is not None:
            data[offset:offset + length] = replace_data[0:length]

        # Byte-swap the data to make it big-endian and return
        data.byteswap()
        return data.tostring()

    @property
    def messages(self):
        """ Get an iterable of message to be sent.
        """

        # Construct and yield the start packet
        yield SpinnakerBootMessage(
            opcode=SpinnakerBootOpCode.FLOOD_FILL_START,
            operand_1=0, operand_2=0, operand_3=self._no_data_packets - 1)

        # Construct and yield the first data packet replacing the appropriate
        # part with the boot data values
        yield SpinnakerBootMessage(
            opcode=SpinnakerBootOpCode.FLOOD_FILL_BLOCK,
            operand_1=_BOOT_DATA_OPERAND_1 | 0, operand_2=0, operand_3=0,
            data=self._get_packet_data(
                self._spinnaker_boot_data, _BOOT_STRUCT_REPLACE_OFFSET,
                _BOOT_STRUCT_REPLACE_LENGTH))

        # Construct an yield the remaining packets
        for block_id in range(1, self._no_data_packets):
            operand_1 = _BOOT_DATA_OPERAND_1 | (block_id & 0xFF)
            yield SpinnakerBootMessage(
                opcode=SpinnakerBootOpCode.FLOOD_FILL_BLOCK,
                operand_1=operand_1, operand_2=0, operand_3=0,
                data=self._get_packet_data())

        # Construct and yield the end packet
        yield SpinnakerBootMessage(
            opcode=SpinnakerBootOpCode.FLOOD_FILL_CONTROL,
            operand_1=1, operand_2=0, operand_3=0)



from spinnman.messages.scp.abstract_messages.abstract_scp_request\
    import AbstractSCPRequest
from spinnman.messages.sdp.sdp_header import SDPHeader
from spinnman.messages.sdp.sdp_flag import SDPFlag
from spinnman.messages.scp.scp_request_header import SCPRequestHeader
from spinnman.messages.scp.enums.scp_command import SCPCommand
from spinnman.messages.scp.enums.scp_dpri_command import SCPDPRICommand
from spinnman.messages.scp.impl.scp_check_ok_response import SCPCheckOKResponse


class SCPDPRISetRouterTimeoutRequest(AbstractSCPRequest):
    """ An SCP Request to set the router timeout for dropped packet reinjection
    """

    def __init__(self, x, y, p, timeout_mantissa, timeout_exponent):
        """
        :param x: The x-coordinate of a chip, between 0 and 255
        :type x: int
        :param y: The y-coordinate of a chip, between 0 and 255
        :type y: int
        :param p: The processor running the dropped packet reinjector, between\
                0 and 17
        :type p: int
        :param timeout_mantissa: The mantissa of the timeout value, \
                between 0 and 15
        :type timeout_mantissa: int
        :param timeout_exponent: The exponent of the timeout value, \
                between 0 and 15
        """
        AbstractSCPRequest.__init__(
            self,
            SDPHeader(
                flags=SDPFlag.REPLY_EXPECTED, destination_port=0,
                destination_cpu=p, destination_chip_x=x,
                destination_chip_y=y),
            SCPRequestHeader(command=SCPCommand.CMD_DPRI),
            argument_1=SCPDPRICommand.SET_ROUTER_TIMEOUT.value,
            argument_2=(timeout_mantissa & 0xF) |
                       ((timeout_exponent & 0xF) << 4))

    def get_scp_response(self):
        return SCPCheckOKResponse("Set router timeout", SCPCommand.CMD_DPRI)



from spinnman.messages.scp.abstract_messages.abstract_scp_request \
    import AbstractSCPRequest
from spinnman.messages.sdp.sdp_header import SDPHeader
from spinnman.messages.sdp.sdp_flag import SDPFlag
from spinnman.messages.scp.scp_request_header import SCPRequestHeader
from spinnman.messages.scp.enums.scp_command import SCPCommand
from spinnman.messages.scp.impl.scp_check_ok_response import SCPCheckOKResponse

_IPTAG_SET = 1


class SCPReverseIPTagSetRequest(AbstractSCPRequest):
    """ An SCP Request to set an IP Tag
    """

    def __init__(self, x, y, destination_x, destination_y, destination_p, port,
                 tag, sdp_port):
        """
        :param x: The x-coordinate of a chip, between 0 and 255
        :type x: int
        :param y: The y-coordinate of a chip, between 0 and 255
        :type y: int
        :param destination_x: The x-coordinate of the destination chip,\
                    between 0 and 255
        :type destination_x: int
        :param destination_y: The y-coordinate of the destination chip,\
                    between 0 and 255
        :type destination_y: int
        :param destination_p: the id of the destination processor, between\
                    0 and 17
        :type destination_p: int
        :param port: The port, between 0 and 65535
        :type port: int
        :param tag: The tag, between 0 and 7
        :type tag: int
        """
        strip_value = 1
        reverse_value = 1

        arg1 = ((reverse_value << 29) | (strip_value << 28) |
                (_IPTAG_SET << 16) | (sdp_port << 13) | (destination_p << 8) |
                tag)
        arg2 = ((destination_x << 24) | (destination_y << 16) | port)

        super(SCPReverseIPTagSetRequest, self).__init__(
            SDPHeader(
                flags=SDPFlag.REPLY_EXPECTED, destination_port=0,
                destination_cpu=0, destination_chip_x=x,
                destination_chip_y=y),
            SCPRequestHeader(command=SCPCommand.CMD_IPTAG),
            argument_1=arg1,
            argument_2=arg2,
            argument_3=0)

    def get_scp_response(self):
        return SCPCheckOKResponse("Set Reverse IP Tag", "CMD_IPTAG")



from six import add_metaclass

from spinn_utilities.abstract_base import AbstractBase, abstractmethod
from spinnman.connections.abstract_classes.abstract_connection \
    import AbstractConnection


@add_metaclass(AbstractBase)
class AbstractMulticastSender(AbstractConnection):
    """ A sender of Multicast messages
    """

    __slots__ = ()

    @abstractmethod
    def get_input_chips(self):
        """ Get a list of chips which identify the chips to which this sender\
            can send multicast packets directly

        :return: An iterable of tuples of (x, y) where x is the x-coordinate\
                    of the chip and y is the y-coordinate of the chip
        :rtype: iterable of (int, int)
        :raise None: No known exceptions are raised
        """
        pass

    @abstractmethod
    def send_multicast_message(self, multicast_message):
        """ Sends a SpiNNaker multicast message using this connection

        :param multicast_message: The message to be sent
        :type multicast_message:\
                    :py:class:`spinnman.messages.multicast_message.MulticastMessage`
        :return: Nothing is returned
        :rtype: None
        :raise spinnman.exceptions.SpinnmanIOException: If there is an error\
                    sending the message
        """
        pass



"""
ScpWriteFPGARegisterRequest
"""

# spinnman imports
import struct

from spinnman.messages.scp.abstract_messages.abstract_scp_bmp_request import \
    AbstractSCPBMPRequest
from spinnman.messages.scp.impl.scp_check_ok_response import SCPCheckOKResponse
from spinnman.messages.scp.enums.scp_command import SCPCommand
from spinnman.messages.scp.scp_request_header import SCPRequestHeader


class SCPWriteFPGARegisterRequest(AbstractSCPBMPRequest):
    """ A request for writing data to a FPGA register
    """

    def __init__(self, fpga_num, addr, value, board):
        """ Write the value of an FPGA (SPI) register.

        See the SpI/O project's spinnaker_fpga design's `README`_ for a listing
        of FPGA registers. The SpI/O project can be found on GitHub at:
        https://github.com/SpiNNakerManchester/spio/

        .. _README: https://github.com/SpiNNakerManchester/spio/\
                    blob/master/designs/spinnaker_fpgas/README.md#spi-interface

        :param fpga_num: FPGA number (0, 1 or 2) to communicate with.
        :type fpga_num: int
        :param addr: Register address to read or write to (will be rounded
                    down to the nearest 32-bit word boundary).
        :type addr: int
        :param value: A 32-bit int value to write to the register
        :type value: int
        """

        AbstractSCPBMPRequest.__init__(
            self, board,
            SCPRequestHeader(command=SCPCommand.CMD_LINK_WRITE),
            argument_1=addr & (~0x3), argument_2=4, argument_3=fpga_num,
            data=struct.pack("<I", value))

    def get_scp_response(self):
        """
        :rtype: SCPCheckOKResponse
        """
        return SCPCheckOKResponse("Send FPGA register write", "CMD_LINK_WRITE")



"""
SCPBMPVersionRequest
"""

# spinnman imports
from spinnman.messages.scp.abstract_messages.abstract_scp_bmp_request import \
    AbstractSCPBMPRequest
from spinnman.messages.scp.scp_request_header import SCPRequestHeader
from spinnman.messages.scp.enums.scp_command import SCPCommand
from spinnman.messages.scp.impl.scp_version_response import SCPVersionResponse


class SCPBMPVersionRequest(AbstractSCPBMPRequest):
    """ An SCP request to read the version of software running on a core
    """

    def __init__(self, board):
        """
        :param board: The board to get the version from
        :type board: int
        :raise spinnman.exceptions.SpinnmanInvalidParameterException:
                    * If the chip coordinates are out of range
                    * If the processor is out of range
        """
        AbstractSCPBMPRequest.__init__(
            self, board,
            SCPRequestHeader(command=SCPCommand.CMD_VER))

    def get_scp_response(self):
        """ See\
            :py:meth:`spinnman.messages.scp.abstract_scp_request.AbstractSCPRequest.get_scp_response`
        """
        return SCPVersionResponse()



from spinnman.messages.scp.impl.scp_dpri_exit_request import SCPDPRIExitRequest
from spinnman.processes.abstract_multi_connection_process \
    import AbstractMultiConnectionProcess


class ExitDPRIProcess(AbstractMultiConnectionProcess):

    def __init__(self, connection_selector):
        AbstractMultiConnectionProcess.__init__(self, connection_selector)

    def exit(self, core_subsets):
        for core_subset in core_subsets.core_subsets:
            for processor_id in core_subset.processor_ids:
                self._send_request(SCPDPRIExitRequest(
                    core_subset.x, core_subset.y, processor_id))
        self._finish()
        self.check_for_error()



import struct


class SCPRequestHeader(object):
    """ Represents the header of an SCP Request
        Each optional parameter in the constructor can be set to a value other\
        than None once, after which it is immutable.  It is an error to set a\
        parameter that is not currently None.
    """

    def __init__(self, command, sequence=0):
        """

        :param command: The SCP command
        :type command: :py:class:`spinnman.messages.scp.scp_command.SCPCommand`
        :param sequence: The number of the SCP packet in order of all packets\
                    sent or received, between 0 and 65535
        :type sequence: int
        :raise spinnman.exceptions.SpinnmanInvalidParameterException: If one\
                    of the parameters is incorrect
        """
        self._command = command
        self._sequence = sequence

    @property
    def command(self):
        """ The command of the SCP packet

        :return: The command
        :rtype: :py:class:`spinnman.messages.scp.scp_command.SCPCommand`
        """
        return self._command

    @property
    def sequence(self):
        """ The sequence number of the SCP packet

        :return: The sequence number of the packet, between 0 and 65535
        :rtype: int
        """
        return self._sequence

    @sequence.setter
    def sequence(self, sequence):
        """ Set the sequence number of the SCP packet

        :param sequence: The sequence number to set, between 0 and 65535
        :type sequence: int
        :return: Nothing is returned
        :rtype: None
        :raise spinnman.exceptions.SpinnmanInvalidParameterException: If the\
                    sequence is out of range, or if it has already been set
        """
        self._sequence = sequence

    @property
    def bytestring(self):
        """ The header as a bytestring

        :return: The header as a bytestring
        :rtype: str
        """
        return struct.pack("<2H", self._command.value, self._sequence)




# spinnman imports
from spinnman.model.bmp_connection_data import BMPConnectionData
from spinnman import constants
from spinnman.messages.sdp.sdp_message import SDPMessage
from spinnman.messages.sdp.sdp_header import SDPHeader
from spinnman.messages.sdp.sdp_flag import SDPFlag
from spinnman.connections.udp_packet_connections import udp_utils

# general imports
import socket


def work_out_bmp_from_machine_details(hostname, number_of_boards):
    """ Work out the BMP connection ip address given the machine details.\
        This is assumed to be the IP address of the machine, with 1 subtracted\
        from the final part e.g. if the machine IP address is 192.168.0.1, the\
        BMP IP address is assumed to be 192.168.0.0

    :param hostname: the spinnaker machine main hostname or IP address
    :param number_of_boards: the number of boards in the machine
    :return: The BMP connection data
    """
    # take the ip address, split by dots, and subtract 1 off last bit
    ipstring = socket.gethostbyname(hostname)
    ip_string_bits = ipstring.split(".")
    ip_string_bits[len(ip_string_bits) - 1] = str(int(
        ip_string_bits[len(ip_string_bits) - 1]) - 1)
    bmp_ip_address_and_port = ".".join(ip_string_bits)
    bmp_bits = bmp_ip_address_and_port.split(",")
    if len(bmp_bits) == 1:
        bmp_ip_address = bmp_bits[0]
        bmp_port = None
    else:
        bmp_ip_address = bmp_bits[0]
        bmp_port = bmp_bits[1]

    # add board scope for each split
    # if None, the end user didn't enter anything, so assume one board
    # starting at position 0
    board_range = list()
    if number_of_boards == 0 or number_of_boards is None:
        board_range.append(int(0))
    else:
        for board_value in range(number_of_boards):
            board_range.append(int(board_value))

    # Assume a single board with no cabinet or frame specified
    return BMPConnectionData(cabinet=0, frame=0, ip_address=bmp_ip_address,
                             boards=board_range, port_num=int(bmp_port))


def get_router_timeout_value(mantissa, exponent):
    """ Get the timeout value of a router in ticks, given the mantissa and\
        exponent of the value

    :param mantissa: The mantissa of the value, between 0 and 15
    :type mantissa: int
    :param exponent: The exponent of the value, between 0 and 15
    :type exponent: int
    """
    if exponent <= 4:
        return ((mantissa + 16) - (2 ** (4 - exponent))) * (2 ** exponent)
    return (mantissa + 16) * (2 ** exponent)


def get_router_timeout_value_from_byte(value):
    """ Get the timeout value of a router in ticks, given an 8-bit floating\
        point value stored in an int(!)

    :param value: The value to convert
    :type value: int
    """
    return get_router_timeout_value(value & 0xF, (value >> 4) & 0XF)


def get_vcpu_address(p):
    """ Get the address of the vcpu_t structure for the given core

    :param p: The core
    :type p: int
    """
    return constants.CPU_INFO_OFFSET + (constants.CPU_INFO_BYTES * p)


def send_port_trigger_message(connection, board_address):
        """ Sends a port trigger message using a connection to (hopefully)\
            open a port in a NAT and/or firewall to allow incoming packets\
            to be received

        :param connection: The UDP connection down which the trigger message\
                should be sent
        :param board_address: The address of the SpiNNaker board to which the\
                message should be sent
        """

        # Set up the message so that no reply is expected and it is sent to an
        # invalid port for SCAMP.  The current version of SCAMP will reject
        # this message, but then fail to send a response since the
        # REPLY_NOT_EXPECTED flag is set (see scamp-3.c line 728 and 625-644)
        trigger_message = SDPMessage(SDPHeader(
            flags=SDPFlag.REPLY_NOT_EXPECTED, tag=0, destination_port=3,
            destination_cpu=0, destination_chip_x=0, destination_chip_y=0))
        udp_utils.update_sdp_header_for_udp_send(
            trigger_message.sdp_header, 0, 0)
        connection.send_to(
            trigger_message.bytestring,
            (board_address, constants.SCP_SCAMP_PORT))



from six import add_metaclass
import struct

from spinn_utilities.abstract_base import AbstractBase, abstractmethod


@add_metaclass(AbstractBase)
class AbstractSCPRequest(object):
    """ Represents an Abstract SCP Request
    """

    DEFAULT_DEST_X_COORD = 255
    DEFAULT_DEST_Y_COORD = 255

    def __init__(self, sdp_header, scp_request_header, argument_1=None,
                 argument_2=None, argument_3=None, data=None):
        """

        :param sdp_header: The SDP header of the request
        :type sdp_header:\
                    :py:class:`spinnman.messages.sdp.sdp_header.SDPHeader`
        :param scp_request_header: The SCP header of the request
        :type scp_request_header:\
                    :py:class:`spinnman.messages.scp.scp_request_header.SCPRequestHeader`
        :param argument_1: The first argument, or None if no first argument
        :type argument_1: int
        :param argument_2: The second argument, or None if no second argument
        :type argument_2: int
        :param argument_3: The third argument, or None if no third argument
        :type argument_3: int
        :param data: The optional data, or None if no data
        :type data: bytearray
        :raise None: No known exceptions are raised
        """
        self._sdp_header = sdp_header
        self._scp_request_header = scp_request_header
        self._argument_1 = argument_1
        self._argument_2 = argument_2
        self._argument_3 = argument_3
        self._data = data

    @property
    def sdp_header(self):
        """ The SDP header of the message

        :rtype: :py:class:`spinnman.message.sdp.sdp_header.SDPHeader`
        """
        return self._sdp_header

    @property
    def scp_request_header(self):
        """ The SCP request header of the message

        :rtype:\
                    :py:class:`spinnman.messages.scp.scp_request_header.SCPRequestHeader`
        """
        return self._scp_request_header

    @property
    def argument_1(self):
        """ The first argument, or None if no first argument

        :rtype: int
        """
        return self._argument_1

    @property
    def argument_2(self):
        """ The second argument, or None if no second argument

        :rtype: int
        """
        return self._argument_2

    @property
    def argument_3(self):
        """ The third argument, or None if no third argument

        :rtype: int
        """
        return self._argument_3

    @property
    def data(self):
        """ The data, or None if no data

        :rtype: bytearray
        """
        return self._data

    @property
    def bytestring(self):
        """ The request as a bytestring

        :return: The request as a bytestring
        :rtype: str
        """
        data = (self._sdp_header.bytestring +
                self._scp_request_header.bytestring)
        if self._argument_1 is not None:
            data += struct.pack("<I", self._argument_1)
        else:
            data += struct.pack("<I", 0)
        if self._argument_2 is not None:
            data += struct.pack("<I", self._argument_2)
        else:
            data += struct.pack("<I", 0)
        if self._argument_3 is not None:
            data += struct.pack("<I", self._argument_3)
        else:
            data += struct.pack("<I", 0)
        if self._data is not None:
            data += bytes(self._data)
        return data

    @abstractmethod
    def get_scp_response(self):
        """ Get an SCP response message to be used to process any response\
            received

        :return: An SCP response, or None if no response is required
        :rtype: :py:class:`spinnman.messages.scp_response.SCPResponse`
        :raise None: No known exceptions are raised
        """
        pass



from enum import Enum


class DiagnosticFilterSource(Enum):
    """ Source flags for the diagnostic filters.
        Note that only one has to match for the counter to be incremented
    """
    LOCAL = (0, "Source is a local core")
    NON_LOCAL = (1, "Source is not a local core")

    def __new__(cls, value, doc=""):
        obj = object.__new__(cls)
        obj._value_ = value
        return obj

    def __init__(self, value, doc=""):
        self._value_ = value
        self.__doc__ = doc



from spinnman.messages.scp.impl.scp_dpri_get_status_request \
    import SCPDPRIGetStatusRequest
from spinnman.processes.abstract_multi_connection_process \
    import AbstractMultiConnectionProcess


class ReadDPRIStatusProcess(AbstractMultiConnectionProcess):

    def __init__(self, connection_selector):
        AbstractMultiConnectionProcess.__init__(self, connection_selector)
        self._dpri_status = None

    def handle_dpri_status_response(self, response):
        self._dpri_status = response.dpri_status

    def get_dpri_status(self, x, y, p):
        self._send_request(SCPDPRIGetStatusRequest(x, y, p),
                           callback=self.handle_dpri_status_response)
        self._finish()
        self.check_for_error()
        return self._dpri_status



from spinnman import exceptions

from spinnman.messages.scp.abstract_messages.abstract_scp_request\
    import AbstractSCPRequest
from spinnman.messages.scp.impl.scp_sdram_alloc_response import \
    SCPSDRAMAllocResponse
from spinnman.messages.sdp.sdp_header import SDPHeader
from spinnman.messages.sdp.sdp_flag import SDPFlag
from spinnman.messages.scp.scp_request_header import SCPRequestHeader
from spinnman.messages.scp.enums.scp_command import SCPCommand
from spinnman.messages.scp.enums.scp_alloc_free_type import SCPAllocFreeType


class SCPSDRAMAllocRequest(AbstractSCPRequest):
    """ An SCP Request to allocate space in the SDRAM space
    """

    def __init__(self, x, y, app_id, size, tag=None):
        """

        :param x: The x-coordinate of the chip to allocate on, between 0 and\
                    255
        :type x: int
        :param y: The y-coordinate of the chip to allocate on, between 0 and\
                    255
        :type y: int
        :param app_id: The id of the application, between 0 and 255
        :type app_id: int
        :param size: The size in bytes of memory to be allocated
        :type size: int
        :param tag: the tag for the SDRAM, a 8-bit (chip-wide) tag that can be\
                looked up by a SpiNNaker application to discover the address\
                of the allocated block. If `0` then no tag is applied.
        :type tag: int
        """

        if tag is None:
            tag = 0
        elif not(0 <= tag < 256):
            raise exceptions.SpinnmanInvalidParameterException(
                "tag",
                "The tag param needs to be between 0 and 255, or None (in "
                "which case 0 will be used by default)", str(tag))

        AbstractSCPRequest.__init__(
            self,
            SDPHeader(
                flags=SDPFlag.REPLY_EXPECTED, destination_port=0,
                destination_cpu=0, destination_chip_x=x,
                destination_chip_y=y),
            SCPRequestHeader(command=SCPCommand.CMD_ALLOC),
            argument_1=(
                (app_id << 8) |
                SCPAllocFreeType.ALLOC_SDRAM.value),  # @UndefinedVariable
            argument_2=size, argument_3=tag)
        self._size = size

    def get_scp_response(self):
        return SCPSDRAMAllocResponse(self._size)



from spinnman.messages.scp.abstract_messages.abstract_scp_request\
    import AbstractSCPRequest
from spinnman.messages.scp.impl.scp_sdram_de_alloc_response import \
    SCPSDRAMDeAllocResponse
from spinnman.messages.sdp.sdp_header import SDPHeader
from spinnman.messages.sdp.sdp_flag import SDPFlag
from spinnman.messages.scp.scp_request_header import SCPRequestHeader
from spinnman.messages.scp.enums.scp_command import SCPCommand
from spinnman.messages.scp.enums.scp_alloc_free_type import SCPAllocFreeType


class SCPSDRAMDeAllocRequest(AbstractSCPRequest):
    """ An SCP Request to free space in the SDRAM
    """

    def __init__(self, x, y, app_id, base_address=None):
        """

        :param x: The x-coordinate of the chip to allocate on, between 0 and\
                    255
        :type x: int
        :param y: The y-coordinate of the chip to allocate on, between 0 and\
                    255
        :type y: int
        :param app_id: The id of the application, between 0 and 255
        :type app_id: int
        :param base_address: The start address in SDRAM to which the block\
                needs to be deallocated, or none if deallocating via app_id
        :type base_address: int or None
        """

        if base_address is not None:
            AbstractSCPRequest.__init__(
                self,
                SDPHeader(
                    flags=SDPFlag.REPLY_NOT_EXPECTED, destination_port=0,
                    destination_cpu=0, destination_chip_x=x,
                    destination_chip_y=y),
                SCPRequestHeader(command=SCPCommand.CMD_ALLOC),
                argument_1=(
                    app_id << 8 |
                    SCPAllocFreeType.
                    FREE_SDRAM_BY_POINTER.value),  # @UndefinedVariable
                argument_2=base_address)
        else:
            AbstractSCPRequest.__init__(
                self,
                SDPHeader(
                    flags=SDPFlag.REPLY_EXPECTED, destination_port=0,
                    destination_cpu=0, destination_chip_x=x,
                    destination_chip_y=y),
                SCPRequestHeader(command=SCPCommand.CMD_ALLOC),
                argument_1=(
                    app_id << 8 |
                    SCPAllocFreeType.
                    FREE_ROUTING_BY_APP_ID.value))  # @UndefinedVariable

    def get_scp_response(self):
        return SCPSDRAMDeAllocResponse()



from enum import Enum


class MailboxCommand(Enum):
    """ Commands sent between an application and the monitor processor
    """

    SHM_IDLE = (0, "The mailbox is idle")
    SHM_MSG = (1, "The mailbox contains an SDP message")
    SHM_NOP = (2, "The mailbox contains a non-operation")
    SHM_SIGNAL = (3, "The mailbox contains a signal")
    SHM_CMD = (4, "The mailbox contains a command")

    def __new__(cls, value, doc=""):
        obj = object.__new__(cls)
        obj._value_ = value
        return obj

    def __init__(self, value, doc=""):
        self._value_ = value
        self.__doc__ = doc



from spinnman.messages.scp.abstract_messages.abstract_scp_request\
    import AbstractSCPRequest
from spinnman.messages.sdp.sdp_header import SDPHeader
from spinnman.messages.sdp.sdp_flag import SDPFlag
from spinnman.messages.scp.scp_request_header import SCPRequestHeader
from spinnman.messages.scp.enums.scp_command import SCPCommand
from spinnman.messages.scp.impl.scp_iptag_get_response \
    import SCPIPTagGetResponse

_IPTAG_GET = 2


class SCPTagGetRequest(AbstractSCPRequest):
    """ An SCP Request to get an IP tag
    """

    def __init__(self, x, y, tag):
        """
        :param x: The x-coordinate of a chip, between 0 and 255
        :type x: int
        :param y: The y-coordinate of a chip, between 0 and 255
        :type y: int
        :param tag: The tag to get details of, between 0 and 7
        :type tag: int
        :param tag: The tag, between 0 and 7
        :type tag: int
        """
        super(SCPTagGetRequest, self).__init__(
            SDPHeader(
                flags=SDPFlag.REPLY_EXPECTED, destination_port=0,
                destination_cpu=0, destination_chip_x=x,
                destination_chip_y=y),
            SCPRequestHeader(command=SCPCommand.CMD_IPTAG),
            argument_1=(_IPTAG_GET << 16) | tag,
            argument_2=1)

    def get_scp_response(self):
        return SCPIPTagGetResponse()



from spinnman.messages.eieio.eieio_type import EIEIOType
from spinnman.messages.eieio.eieio_prefix import EIEIOPrefix
from spinnman.exceptions import SpinnmanInvalidPacketException
import struct


class EIEIODataHeader(object):

    def __init__(self, eieio_type, tag=0, prefix=None,
                 prefix_type=EIEIOPrefix.LOWER_HALF_WORD,
                 payload_base=None, is_time=False, count=0):
        """ EIEIO header for data packets

        :param eieio_type: the type of message
        :type eieio_type:\
                    :py:class:`spinnman.spinnman.messages.eieio.eieio_type.EIEIOType`
        :param tag: the tag of the message (0 by default)
        :type tag: int
        :param prefix: the key prefix of the message or None if not prefixed
        :type prefix: int or None
        :param prefix_type: the position of the prefix (upper or lower)
        :type prefix_type:\
                    :py:class:`spinnman.messages.eieio.eieio_prefix.EIEIOPrefix`
        :param payload_base: The base payload to be applied, or None if no\
                    base payload
        :type payload_base: int or None
        :param is_time: True if the payloads should be taken to be timestamps,\
                    or False otherwise
        :type is_time: bool
        :param count: Count of the number of items in the packet
        :type count: int
        """
        self._eieio_type = eieio_type
        self._tag = tag
        self._prefix = prefix
        self._prefix_type = prefix_type
        self._payload_base = payload_base
        self._is_time = is_time
        self._count = count

    @property
    def eieio_type(self):
        return self._eieio_type

    @property
    def tag(self):
        return self._tag

    @property
    def prefix(self):
        return self._prefix

    @property
    def prefix_type(self):
        return self._prefix_type

    @property
    def payload_base(self):
        return self._payload_base

    @property
    def is_time(self):
        return self._is_time

    @property
    def count(self):
        return self._count

    @count.setter
    def count(self, count):
        self._count = count

    @property
    def size(self):
        return EIEIODataHeader.get_header_size(
            self._eieio_type, self._prefix is not None,
            self._payload_base is not None)

    @staticmethod
    def get_header_size(eieio_type, is_prefix=False, is_payload_base=False):
        """ Get the size of a header with the given parameters

        :param eieio_type: the type of message
        :type eieio_type:\
                    :py:class:`spinnman.spinnman.messages.eieio.eieio_type.EIEIOType`
        :param is_prefix: True if there is a prefix, False otherwise
        :type is_prefix: bool
        :param is_payload_base: True if there is a payload base, False\
                    otherwise
        :type is_payload_base: bool
        :return: The size of the header in bytes
        :rtype: int
        """
        size = 2
        if is_prefix:
            size += 2
        if is_payload_base:
            size += eieio_type.key_bytes
        return size

    def increment_count(self):
        self._count += 1

    def reset_count(self):
        self._count = 0

    @property
    def bytestring(self):
        """ Get a bytestring of the header

        :return: The header as a bytestring
        :rtype: str
        """

        # Convert the flags to an int
        data = 0

        # the flag for prefix or not
        if self._prefix is not None:
            data |= 1 << 7

            # the prefix type
            data |= self._prefix_type.value << 6

        # the flag for payload prefix
        if self._payload_base is not None:
            data |= 1 << 5

        # the flag for time in payloads
        if self._is_time:
            data |= 1 << 4

        # The type of the packet
        data |= self._eieio_type.value << 2

        # The tag of the packet
        data |= self._tag

        # Convert the remaining data, depending on the various options
        if self._payload_base is not None:
            if (self._eieio_type == EIEIOType.KEY_PAYLOAD_16_BIT or
                    self._eieio_type == EIEIOType.KEY_16_BIT):
                if self._prefix is not None:
                    return struct.pack(
                        "<BBHH", self._count, data, self._prefix,
                        self._payload_base)
                else:
                    return struct.pack(
                        "<BBH", self._count, data, self._payload_base)
            elif (self._eieio_type == EIEIOType.KEY_PAYLOAD_32_BIT or
                    self._eieio_type == EIEIOType.KEY_32_BIT):
                if self._prefix is not None:
                    return struct.pack(
                        "<BBHI", self._count, data, self._prefix,
                        self._payload_base)
                else:
                    return struct.pack(
                        "<BBI", self._count, data, self._payload_base)
        else:
            if self._prefix is not None:
                return struct.pack(
                    "<BBH", self._count, data, self._prefix)
            else:
                return struct.pack(
                    "<BB", self._count, data)

    @staticmethod
    def from_bytestring(data, offset):
        """ Read an eieio data header from a bytestring

        :param data: The bytestring to be read
        :type data: str
        :param offset: The offset at which the data starts
        :type offset: int
        :return: an EIEIO header
        :rtype:\
                    :py:class:`spinnman.messages.eieio.data_messages.eieio_data_header.EIEIODataHeader`
        """

        (count, header_data) = struct.unpack_from("<BB", data, offset)

        # Read the flags in the header
        prefix_flag = (header_data >> 7) & 1
        format_flag = (header_data >> 6) & 1
        payload_prefix_flag = (header_data >> 5) & 1
        payload_is_timestamp = (header_data >> 4) & 1
        message_type = (header_data >> 2) & 3
        tag = header_data & 3

        # Check for command header
        if prefix_flag == 0 and format_flag == 1:
            raise SpinnmanInvalidPacketException(
                "EIEIODataHeader",
                "The header indicates that this is a command header")

        # Convert the flags into types
        eieio_type = EIEIOType(message_type)
        prefix_type = EIEIOPrefix(format_flag)

        prefix = None
        payload_prefix = None
        if payload_prefix_flag == 1:
            if (eieio_type == EIEIOType.KEY_16_BIT or
                    eieio_type == EIEIOType.KEY_PAYLOAD_16_BIT):
                if prefix_flag == 1:
                    (prefix, payload_prefix) = struct.unpack_from(
                        "<HH", data, offset + 2)
                else:
                    payload_prefix = struct.unpack_from(
                        "<H", data, offset + 2)[0]
            elif (eieio_type == EIEIOType.KEY_32_BIT or
                    eieio_type == EIEIOType.KEY_PAYLOAD_32_BIT):
                if prefix_flag == 1:
                    (prefix, payload_prefix) = struct.unpack_from(
                        "<HI", data, offset + 2)
                else:
                    payload_prefix = struct.unpack_from(
                        "<I", data, offset + 2)[0]
        else:
            if prefix_flag == 1:
                prefix = struct.unpack_from("<H", data, offset + 2)[0]

        return EIEIODataHeader(
            eieio_type=eieio_type, tag=tag, prefix=prefix,
            prefix_type=prefix_type, payload_base=payload_prefix,
            is_time=bool(payload_is_timestamp), count=count)

    def __str__(self):
        return ("EIEIODataHeader:prefix={}:prefix_type={}:payload_base={}:"
                "is_time={}:type={}:tag={}:count={}".format(
                    self._prefix, self._prefix_type, self._payload_base,
                    self._is_time, self._type.value, self._tag,
                    self._count))

    def __repr__(self):
        return self.__str__()



from spinnman.processes.abstract_process import AbstractProcess
from spinnman.connections.scp_request_pipeline import SCPRequestPipeLine


class AbstractSingleConnectionProcess(AbstractProcess):
    """ A process that uses a single connection in communication
    """

    def __init__(self, connection_selector):
        AbstractProcess.__init__(self)
        self._scp_request_pipeline = None
        self._connection_selector = connection_selector

    def _send_request(self, request, callback=None, error_callback=None):
        if error_callback is None:
            error_callback = self._receive_error

        # If no pipe line built yet, build one on the connection selected for
        # it
        if self._scp_request_pipeline is None:
            self._scp_request_pipeline = SCPRequestPipeLine(
                self._connection_selector.get_next_connection(request))

        # send request
        self._scp_request_pipeline.send_request(
            request, callback, error_callback)

    def _finish(self):
        self._scp_request_pipeline.finish()



from enum import Enum


class SCPPowerCommand(Enum):
    """ The SCP Power Commands
    """

    POWER_OFF = (0, "Power off the machine")
    POWER_ON = (1, "Power on the machine")

    def __new__(cls, value, doc=""):
        obj = object.__new__(cls)
        obj._value_ = value
        return obj

    def __init__(self, value, doc=""):
        self._value_ = value
        self.__doc__ = doc



"""
SCPBMPSetLedRequest
"""
from spinnman.messages.scp.abstract_messages.abstract_scp_bmp_request import \
    AbstractSCPBMPRequest
from spinnman.messages.scp.impl.scp_check_ok_response import SCPCheckOKResponse
from spinnman.messages.scp.enums.scp_command import SCPCommand
from spinnman.messages.scp.scp_request_header import SCPRequestHeader


class SCPBMPSetLedRequest(AbstractSCPBMPRequest):
    """ Set the led(s) of a board to either on, off or toggling
    """

    def __init__(self, led, action, boards):
        """

        :param led: Number of the LED or an iterable of LEDs to set the\
                state of (0-7)
        :type led: int or iterable of int
        :param action: State to set the LED to, either on, off or toggle
        :type action:\
                :py:class:`spinnman.messages.scp.scp_led_action.SCPLEDAction`
        :param boards: Specifies the board to control the LEDs of. This may\
                also be an iterable of multiple boards (in the same frame).
        :type boards: int or iterable of int
        :rtype: None
        """

        # set up the led entry for arg1
        if isinstance(led, int):
            leds = [led]
        else:
            leds = led

        # LED setting actions
        arg1 = sum(action.value << (led * 2) for led in leds)

        # Bitmask of boards to control
        arg2 = self.get_board_mask(boards)

        # initilise the request now
        AbstractSCPBMPRequest.__init__(
            self, boards,
            SCPRequestHeader(command=SCPCommand.CMD_LED),
            argument_1=arg1, argument_2=arg2)

    def get_scp_response(self):
        """ Get the response from the write fpga register request
        """
        return SCPCheckOKResponse("Set the LEDs of a board", "CMD_LED")



from spinnman.messages.scp.impl.scp_application_run_request\
    import SCPApplicationRunRequest
from spinnman.processes.abstract_multi_connection_process \
    import AbstractMultiConnectionProcess


class ApplicationRunProcess(AbstractMultiConnectionProcess):

    def __init__(self, connection_selector):
        AbstractMultiConnectionProcess.__init__(
            self, connection_selector)

    def run(self, app_id, core_subsets, wait):
        for core_subset in core_subsets:
            x = core_subset.x
            y = core_subset.y
            self._send_request(
                SCPApplicationRunRequest(
                    app_id, x, y, core_subset.processor_ids, wait))
        self._finish()
        self.check_for_error()



from spinnman.messages.scp.abstract_messages.abstract_scp_request\
    import AbstractSCPRequest
from spinnman.messages.sdp.sdp_header import SDPHeader
from spinnman.messages.sdp.sdp_flag import SDPFlag
from spinnman.messages.scp.scp_request_header import SCPRequestHeader
from spinnman.messages.scp.enums.scp_command import SCPCommand
from spinnman.messages.scp.impl.scp_check_ok_response import SCPCheckOKResponse

_NNP_FORWARD_RETRY = (0x3f << 8) | 0x18
_NNP_FLOOD_FILL_END = 15
_WAIT_FLAG = 0x1 << 18


class SCPFloodFillEndRequest(AbstractSCPRequest):
    """ A request to start a flood fill of data
    """

    def __init__(
            self, nearest_neighbour_id, app_id=0, processors=None, wait=False):
        """

        :param nearest_neighbour_id: The id of the packet, between 0 and 127
        :type nearest_neighbour_id: int
        :param app_id: The application id to start using the data, between 16\
                    and 255.  If not specified, no application is started
        :type app_id: int
        :param processors: A list of processors on which to start the\
                    application, each between 1 and 17.  If not specified,\
                    no application is started.
        :type processors: iterable of int
        :param wait: True if the binary should go into a "wait" state before\
                    executing
        :type wait: bool
        """
        processor_mask = 0
        if processors is not None:
            for processor in processors:
                processor_mask |= (1 << processor)

        key = (_NNP_FLOOD_FILL_END << 24) | nearest_neighbour_id
        data = (app_id << 24) | processor_mask
        if wait:
            data = data | _WAIT_FLAG

        super(SCPFloodFillEndRequest, self).__init__(
            SDPHeader(flags=SDPFlag.REPLY_EXPECTED, destination_port=0,
                      destination_cpu=0,
                      destination_chip_x=self.DEFAULT_DEST_X_COORD,
                      destination_chip_y=self.DEFAULT_DEST_Y_COORD),
            SCPRequestHeader(command=SCPCommand.CMD_NNP),
            argument_1=key, argument_2=data, argument_3=_NNP_FORWARD_RETRY)

    def get_scp_response(self):
        return SCPCheckOKResponse("Flood Fill", "CMD_NNP:NNP_FFS")



class IOBuffer(object):
    """ The contents of IOBUF for a core
    """

    def __init__(self, x, y, p, iobuf):
        """
        :param x: The x-coordinate of a chip
        :type x: int
        :param y: The y-coordinate of a chip
        :type y: int
        :param p: The p-coordinate of a chip
        :type p: int
        :param iobuf: The contents of the buffer for the chip
        :type iobuf: str
        :raise None: No known exceptions are raised
        """
        self._x = x
        self._y = y
        self._p = p
        self._iobuf = iobuf

    @property
    def x(self):
        """ The x-coordinate of the chip containing the core

        :return: The x-coordinate of the chip
        :rtype: int
        """
        return self._x

    @property
    def y(self):
        """ The y-coordinate of the chip containing the core

        :return: The y-coordinate of the chip
        :rtype: int
        """
        return self._y

    @property
    def p(self):
        """ The id of the core on the chip

        :return: The id of the core
        :rtype: int
        """
        return self._p

    @property
    def iobuf(self):
        """ The contents of the buffer

        :return: The contents of the buffer
        :rtype: str
        """
        return self._iobuf

    def __str__(self):
        value = ""
        for line in self._iobuf.split("\n"):
            value += "{}:{}:{:2n}: {}\n".format(
                self._x, self._y, self._p, line)
        return value[:-1]



from spinn_machine.core_subsets import CoreSubsets


class ExecutableTargets(object):
    """ Encapsulate the binaries and cores on which to execute them
    """

    def __init__(self):
        self._targets = dict()
        self._total_processors = 0
        self._all_core_subsets = CoreSubsets()

    def add_subsets(self, binary, subsets):
        """ Add core subsets to a binary

        :param binary: the path to the binary needed to be executed
        :param subsets: the subset of cores that the binary needs to be loaded\
                    on
        :return:
        """
        if binary in self._targets:
            self._targets[binary].add_core_subset(subsets)
        else:
            self._targets[binary] = subsets

        for subset in subsets.core_subsets:
            for p in subset.processor_ids:
                self._total_processors += 1
                self._all_core_subsets.add_processor(subset.x, subset.y, p)

    def add_processor(self, binary, chip_x, chip_y, chip_p):
        """ Add a processor to the executable targets

        :param binary: the binary path for executable
        :param chip_x: the coordinate on the machine in terms of x for the chip
        :param chip_y: the coordinate on the machine in terms of y for the chip
        :param chip_p: the processor id to place this executable on
        :return:
        """
        if binary not in self._targets:
            self._targets[binary] = CoreSubsets()
        self._targets[binary].add_processor(chip_x, chip_y, chip_p)
        self._all_core_subsets.add_processor(chip_x, chip_y, chip_p)
        self._total_processors += 1

    def get_cores_for_binary(self, binary):
        """ Get the cores that a binary is to run on

        :param binary: The binary to find the cores for
        """
        return self._targets.get(binary)

    @property
    def binaries(self):
        """ The binaries of the executables
        """
        return self._targets.keys()

    @property
    def total_processors(self):
        """ The total number of cores to be loaded
        """
        return self._total_processors

    @property
    def all_core_subsets(self):
        """ All the core subsets for all the binaries
        """
        return self._all_core_subsets



import sys
import time

from spinnman.messages.scp.enums.scp_result import SCPResult
from spinnman.exceptions import SpinnmanTimeoutException

MAX_SEQUENCE = 65536

# Keep a global track of the sequence numbers used
_next_sequence = 0


class SCPRequestPipeLine(object):
    """ Allows a set of SCP requests to be grouped together in a communication\
        across a number of channels for a given connection.

        This class implements an SCP windowing, first suggested by Andrew\
        Mundy.  This extends the idea by having both send and receive windows.\
        These are represented by the n_channels and the\
        intermediate_channel_waits parameters respectively.  This seems to\
        help with the timeout issue; when a timeout is received, all requests\
        for which a reply has not been received can also timeout.
    """

    def __init__(self, connection, n_channels=1,
                 intermediate_channel_waits=0,
                 retry_codes=set([SCPResult.RC_TIMEOUT,
                                  SCPResult.RC_P2P_TIMEOUT,
                                  SCPResult.RC_LEN,
                                  SCPResult.RC_P2P_NOREPLY]),
                 n_retries=3, packet_timeout=0.5):
        """
        :param connection: The connection over which the communication is to\
                    take place
        :param n_channels: The number of requests to send before checking for\
                    responses.  If None, this will be determined automatically
        :param intermediate_channel_waits: The number of outstanding responses\
                    to wait for before continuing sending requests.  If None,\
                    this will be determined automatically
        :param retry_codes: The set of response codes that will be retried
        :param n_retries: The number of times to resend any packet for any\
                    reason before an error is triggered
        :param packet_timeout: The number of elapsed seconds after sending a\
                    packet before it is considered a timeout.
        """
        self._connection = connection
        self._n_channels = n_channels
        self._intermediate_channel_waits = intermediate_channel_waits
        self._retry_codes = retry_codes
        self._n_retries = n_retries
        self._packet_timeout = packet_timeout

        if (self._n_channels is not None and
                self._intermediate_channel_waits is None):
            self._intermediate_channel_waits = self._n_channels - 8
            if self._intermediate_channel_waits < 0:
                self._intermediate_channel_waits = 0

        # A dictionary of sequence number -> requests in progress
        self._requests = dict()
        self._request_data = dict()

        # A dictionary of sequence number -> time at which sequence was sent
        self._times_sent = dict()

        # A dictionary of sequence number -> number of retries for the packet
        self._retries = dict()

        # A dictionary of sequence number -> callback function for response
        self._callbacks = dict()

        # A dictionary of sequence number -> callback function for errors
        self._error_callbacks = dict()

        self._send_time = dict()

        # The number of responses outstanding
        self._in_progress = 0

        # The number of timeouts that occured
        self._n_timeouts = 0

        # The number of packets that have been resent
        self._n_resent = 0
        self._n_retry_code_resent = 0

        # self._token_bucket = TokenBucket(43750, 4375000)
        # self._token_bucket = TokenBucket(3408, 700000)

    def send_request(self, request, callback, error_callback):
        """ Add an SCP request to the set to be sent

        :param request: The SCP request to be sent
        :param callback: A callback function to call when the response has\
                    been received; takes SCPResponse as a parameter, or None\
                    if the response doesn't need to be processed
        :param error_callback: A callback function to call when an error is
                    found when processing the message; takes original\
                    SCPRequest, exception caught and a list of tuples of\
                    (filename, line number, function name, text) as a traceback
        """

        # Update the packet and store required details
        global _next_sequence
        request.scp_request_header.sequence = _next_sequence
        request_data = self._connection.get_scp_data(request)
        self._requests[_next_sequence] = request
        self._request_data[_next_sequence] = request_data
        self._retries[_next_sequence] = self._n_retries
        self._callbacks[_next_sequence] = callback
        self._error_callbacks[_next_sequence] = error_callback
        self._send_time[_next_sequence] = time.time()
        _next_sequence = (_next_sequence + 1) % MAX_SEQUENCE

        # If the connection has not been measured
        if self._n_channels is None:
            if self._connection.is_ready_to_receive():
                self._n_channels = self._in_progress + 8
                if self._n_channels < 12:
                    self._n_channels = 12
                self._intermediate_channel_waits = self._n_channels - 8

        # If all the channels are used, start to receive packets
        while (self._n_channels is not None and
                self._in_progress >= self._n_channels):
            self._do_retrieve(self._intermediate_channel_waits, 0.1)

        # Send the request, keeping track of how many are sent
        # self._token_bucket.consume(284)
        self._connection.send(request_data)
        self._in_progress += 1

    def finish(self):
        """ Indicate the end of the packets to be sent.  This must be called\
            to ensure that all responses are received and handled.
        """
        while self._in_progress > 0:
            self._do_retrieve(0, self._packet_timeout)

    @property
    def n_timeouts(self):
        return self._n_timeouts

    @property
    def n_channels(self):
        return self._n_channels

    @property
    def n_resent(self):
        return self._n_resent

    @property
    def n_retry_code_resent(self):
        return self._n_retry_code_resent

    def _do_retrieve(self, n_packets, timeout):
        """ Receives responses until there are only n_packets responses left

        :param n_packets: The number of packets that can remain after running
        """

        # Keep a set of packets to resend
        to_resend = list()

        # While there are still more packets in progress than some threshold
        while self._in_progress > n_packets:
            try:

                # Receive the next response
                result, seq, raw_data, offset = \
                    self._connection.receive_scp_response(timeout)

                # Only process responses which have matching requests
                if seq in self._requests:
                    request_sent = self._requests[seq]

                    # If the response can be retried, retry it, as long as the
                    # timeout hasn't expired

                    if (result in self._retry_codes and
                            (time.time() - self._send_time[seq] <
                                self._packet_timeout)):
                        self._connection.send(self._request_data[seq])
                        self._n_retry_code_resent += 1
                    else:

                        # No retry is possible - try constructing the result
                        try:
                            response = request_sent.get_scp_response()
                            response.read_bytestring(raw_data, offset)
                            if self._callbacks[seq] is not None:
                                self._callbacks[seq](response)
                        except Exception as e:
                            self._error_callbacks[seq](
                                request_sent, e, sys.exc_info()[2])

                        # Remove the sequence from the outstanding responses
                        del self._send_time[seq]
                        del self._requests[seq]
                        del self._request_data[seq]
                        del self._retries[seq]
                        del self._callbacks[seq]
                        del self._error_callbacks[seq]
                        self._in_progress -= 1
            except SpinnmanTimeoutException:
                self._n_timeouts += 1

                # If there is a timeout, all packets remaining are resent
                time_now = time.time()
                to_remove = list()
                for seq, request_sent in self._requests.iteritems():
                    if time_now - self._send_time[seq] >= self._packet_timeout:
                        to_resend.append((seq, request_sent))
                        self._in_progress -= 1
                        to_remove.append(seq)
                for seq in to_remove:
                    del self._requests[seq]

        # Try to resend the packets
        for seq, request_sent in to_resend:
            if self._retries[seq] > 0:

                # If the request can be retried, retry it
                self._retries[seq] -= 1
                self._in_progress += 1
                self._requests[seq] = request_sent
                self._send_time[seq] = time.time()
                self._connection.send(self._request_data[seq])
                self._n_resent += 1
            else:

                # Otherwise, report it as a timeout
                try:
                    raise SpinnmanTimeoutException(
                        request_sent.scp_request_header.command,
                        self._packet_timeout)
                except Exception as e:
                    self._error_callbacks[seq](
                        request_sent, e, sys.exc_info()[2])
                    del self._request_data[seq]
                    del self._send_time[seq]
                    del self._retries[seq]
                    del self._callbacks[seq]
                    del self._error_callbacks[seq]



from spinnman.messages.scp.impl.scp_write_link_request \
    import SCPWriteLinkRequest
from spinnman.processes.abstract_multi_connection_process \
    import AbstractMultiConnectionProcess
from spinnman.messages.scp.impl.scp_write_memory_request \
    import SCPWriteMemoryRequest
from spinnman import constants

import functools


class WriteMemoryProcess(AbstractMultiConnectionProcess):
    """ A process for writing memory
    """

    def __init__(self, connection_selector):
        AbstractMultiConnectionProcess.__init__(
            self, connection_selector,
            n_channels=1, intermediate_channel_waits=0)

    def write_memory_from_bytearray(
            self, x, y, p, base_address, data, offset, n_bytes):
        """
        writes memory onto a spinnaker chip from a bytearray

        :param x: the x coord of the chip in question
        :param y: the y coord of the chip in question
        :param p: the p coord of the chip in question
        :param base_address: the address in sdram to start writing
        :param data: the data to write
        :param offset: where in the data to start writing from
        :param n_bytes: how much data to write
        :rtype: None
        """
        self._write_memory_from_bytearray(
            base_address, data, offset, n_bytes,
            functools.partial(SCPWriteMemoryRequest, x=x, y=y, cpu=p))

    def write_link_memory_from_bytearray(self, x, y, p, link, base_address,
                                         data, offset, n_bytes):
        self._write_memory_from_bytearray(
            base_address, data, offset, n_bytes,
            functools.partial(SCPWriteLinkRequest, x=x, y=y, cpu=p, link=link))

    def write_memory_from_reader(self, x, y, p, base_address, data, n_bytes):
        self._write_memory_from_reader(
            base_address, data, n_bytes,
            functools.partial(SCPWriteMemoryRequest, x=x, y=y, cpu=p))

    def write_link_memory_from_reader(self, x, y, p, link, base_address, data,
                                      n_bytes):
        self._write_memory_from_reader(
            base_address, data, n_bytes,
            functools.partial(SCPWriteLinkRequest, x=x, y=y, cpu=p, link=link))

    def _write_memory_from_bytearray(self, base_address, data, offset,
                                     n_bytes, packet_class):
        data_offset = offset
        offset = base_address
        n_bytes_to_write = int(n_bytes)
        while n_bytes_to_write > 0:

            bytes_to_send = n_bytes_to_write
            if bytes_to_send > constants.UDP_MESSAGE_MAX_SIZE:
                bytes_to_send = constants.UDP_MESSAGE_MAX_SIZE

            request = packet_class(
                base_address=offset,
                data=data[data_offset:data_offset + bytes_to_send])
            self._send_request(request)

            n_bytes_to_write -= bytes_to_send
            offset += bytes_to_send
            data_offset += bytes_to_send
        self._finish()
        self.check_for_error()

    def _write_memory_from_reader(self, base_address, data, n_bytes,
                                  packet_class):
        offset = base_address
        n_bytes_to_write = int(n_bytes)
        while n_bytes_to_write > 0:

            bytes_to_send = n_bytes_to_write
            if bytes_to_send > constants.UDP_MESSAGE_MAX_SIZE:
                bytes_to_send = constants.UDP_MESSAGE_MAX_SIZE
            data_array = data.read(bytes_to_send)
            data_length = len(data_array)

            request = packet_class(base_address=offset, data=data_array)
            self._send_request(request)

            n_bytes_to_write -= data_length
            offset += data_length
        self._finish()
        self.check_for_error()



import struct

from spinnman.messages.scp.abstract_messages.abstract_scp_response\
    import AbstractSCPResponse
from spinnman.messages.scp.enums.scp_result import SCPResult
from spinnman.exceptions import SpinnmanUnexpectedResponseCodeException


class SCPIPTagGetResponse(AbstractSCPResponse):
    """ An SCP response to a request for an IP tags
    """

    def __init__(self):
        """
        """
        super(SCPIPTagGetResponse, self).__init__()
        self._ip_address = None
        self._mac_address = None
        self._port = None
        self._timeout = None
        self._flags = None
        self._count = None
        self._rx_port = None
        self._spin_chip_y = None
        self._spin_chip_x = None
        self._spin_port = None

    def read_data_bytestring(self, data, offset):
        """ See\
            :py:meth:`spinnman.messages.scp.abstract_scp_response.AbstractSCPResponse.read_data_bytestring`
        """
        result = self.scp_response_header.result
        if result != SCPResult.RC_OK:
            raise SpinnmanUnexpectedResponseCodeException(
                "Get IP Tag Info", "CMD_IPTAG", result.name)
        (ip_address, mac_address, self._port, self._timeout,
         self._flags, self._count, self._rx_port, self._spin_chip_y,
         self._spin_chip_x, processor_and_port) = struct.unpack_from(
            "<4s6s3HIH3B", data, offset)
        self._ip_address = bytearray(ip_address)
        self._mac_address = bytearray(mac_address)
        self._spin_port = (processor_and_port >> 5) & 0x7
        self._spin_cpu = processor_and_port & 0x1F

    @property
    def ip_address(self):
        """ The IP address of the tag, as a bytearray of 4 bytes

        :rtype: bytearray
        """
        return self._ip_address

    @property
    def mac_address(self):
        """ The MAC address of the tag, as a bytearray of 6 bytes

        :rtype: bytearray
        """
        return self._mac_address

    @property
    def port(self):
        """ The port of the tag

        :rtype: int
        """
        return self._port

    @property
    def timeout(self):
        """ The timeout of the tag

        :rtype: int
        """
        return self._timeout

    @property
    def flags(self):
        """ The flags of the tag

        :rtype: int
        """
        return self._flags

    @property
    def in_use(self):
        """ True if the tag is marked as being in use

        :rtype: bool
        """
        return (self._flags & 0x8000) > 0

    @property
    def is_temporary(self):
        """ True if the tag is temporary

        :rtype: bool
        """
        return (self._flags & 0x4000) > 0

    @property
    def is_arp(self):
        """ True if the tag is in the ARP state (where the MAC address is\
            being looked up - transient state so unlikely)

        :rtype: bool
        """
        return (self._flags & 0x2000) > 0

    @property
    def is_reverse(self):
        """ True if the tag is a reverse tag

        :rtype: bool
        """
        return (self._flags & 0x0200) > 0

    @property
    def strip_sdp(self):
        """ True if the tag is to strip sdp

        :rtype: bool
        """
        return (self._flags & 0x0100) > 0

    @property
    def count(self):
        """ The count of the number of packets that have been sent with the tag

        :rtype: int
        """
        return self._count

    @property
    def rx_port(self):
        """ The receive port of the tag

        :rtype: int
        """
        return self._rx_port

    @property
    def spin_chip_x(self):
        """ The x-coordinate of the chip on which the tag is defined

        :rtype: int
        """
        return self._spin_chip_x

    @property
    def spin_chip_y(self):
        """ The y-coordinate of the chip on which the tag is defined

        :rtype: int
        """
        return self._spin_chip_y

    @property
    def spin_port(self):
        """ The spin-port of the ip tag

        :rtype: int
        """
        return self._spin_port

    @property
    def spin_cpu(self):
        """ The cpu id of the ip tag

        :rtype: int
        """
        return self._spin_cpu



from spinnman.processes.abstract_multi_connection_process_connection_selector \
    import AbstractMultiConnectionProcessConnectionSelector


class MultiConnectionProcessRoundRobinConnectionSelector(
        AbstractMultiConnectionProcessConnectionSelector):
    """
    MultiConnectionProcessRoundRobinConnectionSelector: selector
    """

    def __init__(self, connections):
        AbstractMultiConnectionProcessConnectionSelector.__init__(
            self, connections)
        self._connections = connections
        self._next_connection_index = 0

    def get_next_connection(self, message):
        index = self._next_connection_index
        self._next_connection_index = (index + 1) % len(self._connections)
        return self._connections[index]



from enum import Enum


class DiagnosticFilterEmergencyRoutingStatus(Enum):
    """ Emergency routing status flags for the diagnostic filters.
        Note that only one has to match for the counter to be incremented
    """
    NORMAL = (0, "Packet is not emergency routed")
    FIRST_STAGE_COMBINED = (1, "Packet is in first hop of emergency route;"
                               " packet should also have been sent here by"
                               " normal routing")
    FIRST_STAGE = (2, "Packet is in first hop of emergency route; packet"
                      " wouldn't have reached this router without emergency"
                      " routing")
    SECOND_STAGE = (3, "Packet is in last hop of emergency route and should"
                       " now return to normal routing")

    def __new__(cls, value, doc=""):
        obj = object.__new__(cls)
        obj._value_ = value
        return obj

    def __init__(self, value, doc=""):
        self._value_ = value
        self.__doc__ = doc



from enum import Enum


class SignalType(Enum):
    """ The type of signal, determined by how it is transmitted
    """
    MULTICAST = 0
    POINT_TO_POINT = 1
    NEAREST_NEIGHBOUR = 2


class SCPSignal(Enum):
    """ SCP Signals
    """
    INITIALISE = (0, SignalType.NEAREST_NEIGHBOUR)
    POWER_DOWN = (1, SignalType.NEAREST_NEIGHBOUR)
    STOP = (2, SignalType.NEAREST_NEIGHBOUR)
    START = (3, SignalType.NEAREST_NEIGHBOUR)
    SYNC0 = (4, SignalType.MULTICAST)
    SYNC1 = (5, SignalType.MULTICAST)
    PAUSE = (6, SignalType.MULTICAST)
    CONTINUE = (7, SignalType.MULTICAST)
    EXIT = (8, SignalType.MULTICAST)
    TIMER = (9, SignalType.MULTICAST)
    USER_0 = (10, SignalType.MULTICAST)
    USER_1 = (11, SignalType.MULTICAST)
    USER_2 = (12, SignalType.MULTICAST)
    USER_3 = (13, SignalType.MULTICAST)

    def __new__(cls, value, signal_type, doc=""):
        obj = object.__new__(cls)
        obj._value_ = value
        return obj

    def __init__(self, value, signal_type, doc=""):
        """

        :param value: The value used for the signal
        :type value: int
        :param signal_type: The "type" of the signal
        :type signal_type: :py:class:`.SignalType`
        """
        self._value_ = value
        self._signal_type = signal_type
        self.__doc__ = doc

    @property
    def signal_type(self):
        return self._signal_type



from six import add_metaclass

from spinn_utilities.abstract_base import AbstractBase, abstractmethod


@add_metaclass(AbstractBase)
class AbstractListenable(object):

    __slots__ = ()

    @abstractmethod
    def get_receive_method(self):
        """ Get the method that receives for this connection
        """

    @abstractmethod
    def is_ready_to_receive(self, timeout=0):
        """ Determines if there is an SCP packet to be read without blocking

        :param timeout: The time to wait before returning if the connection\
                is not ready
        :type timeout: int
        :return: True if there is an SCP packet to be read
        :rtype: bool
        """
        pass



from spinnman.processes.abstract_multi_connection_process\
    import AbstractMultiConnectionProcess
from spinnman.messages.scp.impl.scp_iptag_info_request import SCPTagInfoRequest
from spinnman.messages.scp.impl.scp_iptag_get_request import SCPTagGetRequest

from spinn_machine.tags.reverse_iptag import ReverseIPTag
from spinn_machine.tags.iptag import IPTag

import functools


class GetTagsProcess(AbstractMultiConnectionProcess):

    def __init__(self, connection_selector):
        AbstractMultiConnectionProcess.__init__(self, connection_selector)
        self._tag_info = None
        self._tags = None

    def handle_tag_info_response(self, response):
        self._tag_info = response

    def handle_get_tag_response(self, tag, board_address, response):
        if response.in_use:
            ip_address = response.ip_address
            host = "{}.{}.{}.{}".format(ip_address[0], ip_address[1],
                                        ip_address[2], ip_address[3])
            if response.is_reverse:
                self._tags[tag] = ReverseIPTag(
                    board_address, tag,
                    response.rx_port, response.spin_chip_x,
                    response.spin_chip_y, response.spin_cpu,
                    response.spin_port)
            else:
                self._tags[tag] = IPTag(
                    board_address, response.sdp_header.source_chip_x,
                    response.sdp_header.source_chip_y, tag, host,
                    response.port, response.strip_sdp)

    def get_tags(self, connection):

        # Get the tag information, without which we cannot continue
        self._send_request(SCPTagInfoRequest(
            connection.chip_x, connection.chip_y),
            self.handle_tag_info_response)
        self._finish()
        self.check_for_error()

        # Get the tags themselves
        n_tags = self._tag_info.pool_size + self._tag_info.fixed_size
        self._tags = [None] * n_tags
        for tag in xrange(n_tags):
            self._send_request(SCPTagGetRequest(
                connection.chip_x, connection.chip_y, tag),
                functools.partial(
                    self.handle_get_tag_response, tag,
                    connection.remote_ip_address))
        self._finish()
        self.check_for_error()

        # Return the tags
        return [tag for tag in self._tags if tag is not None]



from spinnman.messages.scp.abstract_messages.abstract_scp_request\
    import AbstractSCPRequest
from spinnman.exceptions import SpinnmanInvalidParameterException
from spinnman.messages.sdp.sdp_header import SDPHeader
from spinnman.messages.sdp.sdp_flag import SDPFlag
from spinnman.messages.scp.scp_request_header import SCPRequestHeader
from spinnman.messages.scp.enums.scp_command import SCPCommand
from spinnman.messages.scp.impl.scp_check_ok_response import SCPCheckOKResponse


class SCPRouterInitRequest(AbstractSCPRequest):
    """ A request to initialize the router on a chip
    """

    def __init__(self, x, y, n_entries, table_address, base_address, app_id):
        """

        :param x: The x-coordinate of the chip, between 0 and 255, \
        this is not checked due to speed restrictions
        :type x: int
        :param y: The y-coordinate of the chip, between 0 and 255\
        this is not checked due to speed restrictions
        :type y: int
        :param n_entries: The number of entries in the table, more than 0
        :type n_entries: int
        :param table_address: The allocated table address
        :type table_address: int
        :param base_address: The base_address containing the entries
        :type base_address: int
        :param app_id: The id of the application with which to associate the\
                    routes.  If not specified, defaults to 0.
        :type app_id: int
        :raise spinnman.exceptions.SpinnmanInvalidParameterException:\
                    * If x is out of range
                    * If y is out of range
                    * If n_entries is 0 or less
                    * If table_address is not positive
                    * If base_address is not positive
        """
        if n_entries < 1:
            raise SpinnmanInvalidParameterException(
                "n_entries", str(n_entries),
                "Must be more than 0")
        if base_address < 0:
            raise SpinnmanInvalidParameterException(
                "base_address", str(base_address),
                "Must be a positive integer")
        if table_address < 0:
            raise SpinnmanInvalidParameterException(
                "table_address", str(table_address),
                "Must be a positive integer")

        super(SCPRouterInitRequest, self).__init__(
            SDPHeader(
                flags=SDPFlag.REPLY_EXPECTED, destination_port=0,
                destination_cpu=0, destination_chip_x=x,
                destination_chip_y=y),
            SCPRequestHeader(command=SCPCommand.CMD_RTR),
            argument_1=((n_entries << 16) | (app_id << 8) | 2),
            argument_2=table_address, argument_3=base_address)

    def get_scp_response(self):
        return SCPCheckOKResponse("RouterInit", "CMD_RTR")



from six import add_metaclass

from spinn_utilities.abstract_base import AbstractBase, abstractmethod


@add_metaclass(AbstractBase)
class AbstractMultiConnectionProcessConnectionSelector(object):
    """ A connection selector for multi-connection processes
    """

    # connections will be used when worked out how

    @abstractmethod
    def __init__(self, connections):
        """
        :param connections: The connections to be used
        """

    @abstractmethod
    def get_next_connection(self, message):
        """ Get the index of the  next connection for the process from a list\
            of connections

        :param message: The SCP message to be sent
        :rtype: int
        """



import struct

from spinnman.model.enums.cpu_state import CPUState


class ChipSummaryInfo(object):
    """ Represents the chip summary information read via an SCP command
    """

    def __init__(self, chip_summary_data, offset, x, y):
        """
        :param chip_summary_data: The data from the SCP response
        :type chip_summary_data: bytearray
        :param offset: The offset into the data where the data starts
        :type offset: int
        :param x: The x-coordinate of the chip that this data is from
        :param y: The y-coordinate of the chip that this data is from
        """
        (chip_summary_flags, self._largest_free_sdram_block,
            self._largest_free_sram_block) = struct.unpack_from(
                "<3I", chip_summary_data, offset)
        self._n_cores = chip_summary_flags & 0x1F
        self._working_links = [
            link for link in range(0, 6)
            if chip_summary_flags >> (8 + link) & 1 != 0]
        self._n_free_multicast_routing_entries = \
            (chip_summary_flags >> 14) & 0x7FF
        self._is_ethernet_available = bool(chip_summary_flags & (1 << 25))

        data_offset = offset + 12
        self._core_states = [
            CPUState(state) for state in
            struct.unpack_from("<18B", chip_summary_data, data_offset)]
        data_offset += 18

        self._x = x
        self._y = y

        self._nearest_ethernet_x = None
        self._nearest_ethernet_y = None
        self._ethernet_ip_address = None

        (self._nearest_ethernet_y, self._nearest_ethernet_x) = \
            struct.unpack_from(
                "<BB", chip_summary_data, data_offset)
        data_offset += 2

        ip_data = struct.unpack_from(
            "<4B", chip_summary_data, data_offset)
        ethernet_ip_address = "{}.{}.{}.{}".format(
            ip_data[0], ip_data[1], ip_data[2], ip_data[3])
        if self._is_ethernet_available:
            self._ethernet_ip_address = ethernet_ip_address
        data_offset += 4

    @property
    def x(self):
        """ The x-coordinate of the chip that this data is from

        :rtype: int
        """
        return self._x

    @property
    def y(self):
        """ The y-coordinate of the chip that this data is from

        :rtype: int
        """
        return self._y

    @property
    def n_cores(self):
        """ The number of cores working on the chip (including monitors)

        :rtype: int
        """
        return self._n_cores

    @property
    def core_states(self):
        """ The state of the cores on the chip (list of one per core)

        :rtype: list of `py:class:spinnman.model.cpu_state.CPUState`
        """
        return self._core_states

    @property
    def working_links(self):
        """ The ids of the working links outgoing from this chip

        :rtype: list of int
        """
        return self._working_links

    @property
    def is_ethernet_available(self):
        """ Determines if the Ethernet connection is available on this chip

        :rtype: bool
        """
        return self._is_ethernet_available

    @property
    def n_free_multicast_routing_entries(self):
        """ The number of multicast routing entries free on this chip

        :rtype: int
        """
        return self._n_free_multicast_routing_entries

    @property
    def largest_free_sdram_block(self):
        """ The size of the largest block of free SDRAM in bytes

        :rtype: int
        """
        return self._largest_free_sdram_block

    @property
    def largest_free_sram_block(self):
        """ The size of the largest block of free SRAM in bytes

        :rtype: int
        """
        return self._largest_free_sram_block

    @property
    def nearest_ethernet_x(self):
        """ The x coordinate of the nearest Ethernet chip

        :rtype: int
        """
        return self._nearest_ethernet_x

    @property
    def nearest_ethernet_y(self):
        """ The y coordinate of the nearest Ethernet chip

        :rtype: int
        """
        return self._nearest_ethernet_y

    @property
    def ethernet_ip_address(self):
        """ The IP address of the Ethernet if up, or None if not

        :rtype: str
        """
        return self._ethernet_ip_address



from spinnman.messages.scp.abstract_messages.abstract_scp_request\
    import AbstractSCPRequest
from spinnman.messages.sdp.sdp_header import SDPHeader
from spinnman.messages.sdp.sdp_flag import SDPFlag
from spinnman.messages.scp.scp_request_header import SCPRequestHeader
from spinnman.messages.scp.enums.scp_command import SCPCommand
from spinnman.messages.scp.impl.scp_check_ok_response import SCPCheckOKResponse


class SCPWriteLinkRequest(AbstractSCPRequest):
    """ A request to write memory on a neighbouring chip
    """

    def __init__(self, x, y, link, base_address, data, cpu=0):
        """

        :param x: The x-coordinate of the chip whose neighbour will be written\
                    to, between 0 and 255
        :type x: int
        :param y: The y-coordinate of the chip whose neighbour will be written\
                    to, between 0 and 255
        :type y: int
        :param cpu: The CPU core to use, normally 0 (or if a BMP, the board \
                      slot number)
        :type cpu: int
        :param link: The link number to write to between 0 and 5 (or if a BMP,\
                       the FPGA between 0 and 2)
        :type link: int
        :param base_address: The base_address to start writing to
        :type base_address: int
        :param data: Up to 256 bytes of data to write
        :type data: bytearray
        """
        super(SCPWriteLinkRequest, self).__init__(
            SDPHeader(
                flags=SDPFlag.REPLY_EXPECTED, destination_port=0,
                destination_cpu=cpu, destination_chip_x=x,
                destination_chip_y=y),
            SCPRequestHeader(command=SCPCommand.CMD_LINK_WRITE),
            argument_1=base_address, argument_2=len(data), argument_3=link,
            data=None)
        self._data_to_write = data

    @property
    def bytestring(self):
        datastring = super(SCPWriteLinkRequest, self).bytestring
        return datastring + bytes(self._data_to_write)

    def get_scp_response(self):
        return SCPCheckOKResponse("WriteMemory", "CMD_WRITE")
